from datetime import datetime, timezone
import json
import tempfile
from typing import Iterator, Optional, Tuple
from unittest.mock import ANY, call, MagicMock, Mock, mock_open, patch

import click
import pytest
import yaml

from anyscale.client.openapi_client import (
    AppConfig,
    Build,
    Cloud,
    ClusterConfig,
    ComputeTemplateConfig,
    CreateComputeTemplate,
    Project,
    ProjectDefaultSessionName,
    ProjectdefaultsessionnameResponse,
    ProjectListResponse,
    Session,
    SessionListResponse,
    SessionResponse,
    SessionUpOptions,
    SetupInitializeSessionOptions,
    StartEmptySessionResponse,
    StartemptysessionresponseResponse,
    StopSessionOptions,
)
from anyscale.controllers.session_controller import SessionController
from anyscale.sdk.anyscale_client import (
    CreateSession,
    StartSessionOptions,
    UpdateSession,
)
from anyscale.shared_anyscale_utils.project import CLUSTER_YAML_TEMPLATE


COMPUTE_CONFIG_TEMPLATE = json.dumps(
    {
        "cloud_id": "fake-cloud-id",
        "region": "fake-region",
        "allowed_azs": ["fake-az1"],
        "head_node_type": {
            "name": "head-node-name",
            "instance_type": "fake-head-instance-type",
        },
        "worker_node_types": [
            {
                "name": "worker-node-name",
                "instance_type": "fake-worker-instance-type",
                "min_workers": 0,
                "max_workers": 10,
                "use_spot": True,
            }
        ],
        "aws": {
            "SubnetId": "fake-subnet-id",
            "SecurityGroupIds": ["fake-security-group-id"],
            "IamInstanceProfile": "fake-iam-arn",
            "TagSpecifications": [
                {
                    "ResourceType": "instance",
                    "Tags": [{"Key": "fake-key", "Value": "fake-value"}],
                },
            ],
        },
    }
)


@pytest.fixture()
def mock_auth_api_client(mock_api_client: Mock, base_mock_anyscale_api_client: Mock):
    mock_auth_api_client = Mock(
        api_client=mock_api_client, anyscale_api_client=base_mock_anyscale_api_client,
    )
    with patch.multiple(
        "anyscale.controllers.base_controller",
        get_auth_api_client=Mock(return_value=mock_auth_api_client),
    ):
        yield


@pytest.fixture()
def mock_auth_api_client_multiple_sessions(
    mock_api_client_multiple_sessions: Mock, base_mock_anyscale_api_client: Mock
):
    mock_auth_api_client = Mock(
        api_client=mock_api_client_multiple_sessions,
        anyscale_api_client=base_mock_anyscale_api_client,
    )
    with patch.multiple(
        "anyscale.controllers.base_controller",
        get_auth_api_client=Mock(return_value=mock_auth_api_client),
    ):
        yield


@pytest.fixture()
def mock_auth_api_client_with_session(
    mock_api_client_with_session: Mock, base_mock_anyscale_api_client: Mock
):
    mock_auth_api_client = Mock(
        api_client=mock_api_client_with_session,
        anyscale_api_client=base_mock_anyscale_api_client,
    )
    with patch.multiple(
        "anyscale.controllers.base_controller",
        get_auth_api_client=Mock(return_value=mock_auth_api_client),
    ):
        yield


@pytest.fixture()
def mock_api_client(mock_api_client_with_session: Mock) -> Mock:
    mock_api_client = mock_api_client_with_session
    mock_api_client.stop_session_api_v2_sessions_session_id_stop_post.return_value = (
        None
    )

    return mock_api_client


def test_stop_single_session(session_test_data: Session, mock_auth_api_client) -> None:
    session_controller = SessionController()

    mock_project_definition = Mock()
    mock_project_definition.root = "/some/directory"
    mock_load_project_or_throw = Mock(return_value=mock_project_definition)

    mock_get_project_id = Mock(return_value=session_test_data.project_id)

    with patch.multiple(
        "anyscale.controllers.session_controller",
        load_project_or_throw=mock_load_project_or_throw,
        get_project_id=mock_get_project_id,
    ):
        session_controller.stop(
            "test-name", delete=False, workers_only=False, keep_min_workers=False,
        )

    session_controller.api_client.stop_session_api_v2_sessions_session_id_stop_post.assert_called_once_with(
        session_test_data.id,
        StopSessionOptions(
            terminate=True, delete=False, workers_only=False, keep_min_workers=False
        ),
    )

    mock_load_project_or_throw.assert_called_once_with()
    mock_get_project_id.assert_called_once_with("/some/directory")


@pytest.fixture()
def mock_api_client_multiple_sessions(base_mock_api_client: Mock) -> Mock:
    base_mock_api_client.list_sessions_api_v2_sessions_get.return_value = SessionListResponse(
        results=[
            Session(
                id="ses_1",
                name="session_name",
                created_at=datetime.now(tz=timezone.utc),
                snapshots_history=[],
                tensorboard_available=False,
                project_id="project_id",
                state="Running",
                head_node_ip="127.0.0.1",
                idle_timeout=120,
                access_token="value",
            ),
            Session(
                id="ses_2",
                name="session_name2",
                created_at=datetime.now(tz=timezone.utc),
                snapshots_history=[],
                tensorboard_available=False,
                project_id="project_id",
                state="Running",
                head_node_ip="127.0.0.1",
                idle_timeout=120,
                access_token="value",
            ),
        ]
    )
    base_mock_api_client.stop_session_api_v2_sessions_session_id_stop_post.return_value = (
        None
    )
    base_mock_api_client.get_session_head_ip_api_v2_sessions_session_id_head_ip_get.return_value.result.head_ip = (
        "127.0.0.1"
    )

    return base_mock_api_client


def test_stop_multiple_sessions(mock_auth_api_client_multiple_sessions) -> None:
    session_controller = SessionController()

    mock_project_definition = Mock()
    mock_project_definition.root = "/some/directory"
    mock_load_project_or_throw = Mock(return_value=mock_project_definition)

    mock_get_project_id = Mock(return_value="project_id")

    with patch.multiple(
        "anyscale.controllers.session_controller",
        load_project_or_throw=mock_load_project_or_throw,
        get_project_id=mock_get_project_id,
    ):
        session_controller.stop(
            "ses_?", delete=True, workers_only=True, keep_min_workers=True,
        )

    session_controller.api_client.stop_session_api_v2_sessions_session_id_stop_post.assert_has_calls(
        [
            call(
                "ses_1",
                StopSessionOptions(
                    terminate=True,
                    delete=True,
                    workers_only=True,
                    keep_min_workers=True,
                ),
            ),
            call(
                "ses_2",
                StopSessionOptions(
                    terminate=True,
                    delete=True,
                    workers_only=True,
                    keep_min_workers=True,
                ),
            ),
        ]
    )

    mock_load_project_or_throw.assert_called_once_with()
    mock_get_project_id.assert_called_once_with("/some/directory")


@pytest.fixture()
def mock_up_tuple(
    base_mock_api_client: Mock,
    project_test_data: Project,
    cloud_test_data: Cloud,
    session_test_data: Session,
) -> Iterator[Tuple[Mock, Project, Cloud, Mock]]:
    base_mock_api_client.get_session_api_v2_sessions_session_id_get = Mock(
        return_value=SessionResponse(result=session_test_data)
    )
    mock_project_definition = Mock()
    mock_project_definition.root = "/some/directory"
    mock_wait_for_session_start = Mock()

    with patch.multiple(
        "anyscale.controllers.session_controller",
        load_project_or_throw=Mock(return_value=mock_project_definition),
        get_project_id=Mock(return_value=project_test_data.id),
        get_cloud_id_and_name=Mock(
            return_value=tuple([cloud_test_data.id, cloud_test_data.name])
        ),
        wait_for_session_start=mock_wait_for_session_start,
        wait_for_head_node_start=mock_wait_for_session_start,
        _update_file_mounts=MagicMock(
            side_effect=lambda cluster_config, project_root_dir, api_client: cluster_config
        ),
        get_working_dir=Mock(return_value="mock_working_dir"),
    ):
        yield base_mock_api_client, project_test_data, cloud_test_data, mock_wait_for_session_start


def test_anyscale_up_success(
    mock_up_tuple: Tuple[Mock, Project, Cloud, Mock], mock_auth_api_client
) -> None:
    (mock_api_client, project_test_data, cloud_test_data, _,) = mock_up_tuple
    session_controller = SessionController()
    session_controller.push = Mock()  # type: ignore
    mock_api_client.get_project_default_session_name_api_v2_projects_project_id_default_session_name_get.return_value = ProjectdefaultsessionnameResponse(
        result=ProjectDefaultSessionName(name="session-11")
    )
    mock_api_client.list_sessions_api_v2_sessions_get.return_value = SessionListResponse(
        results=[]  # No session with this name yet, brand new session.
    )
    # Session Name should be "11"
    mock_api_client.start_empty_session_api_v2_sessions_start_empty_session_post = Mock(
        return_value=StartemptysessionresponseResponse(
            result=StartEmptySessionResponse(
                session_id="session-10", second_update_required=False,
            )
        )
    )
    with tempfile.NamedTemporaryFile(mode="w") as temp_file:
        temp_file.write(CLUSTER_YAML_TEMPLATE)
        temp_file.flush()

        session_controller.up(
            session_name=None,
            config=temp_file.name,
            build_id=None,
            min_workers=None,
            max_workers=None,
            no_restart=False,
            restart_only=False,
            disable_sync=False,
            cloud_id=None,
            cloud_name=None,
            idle_timeout=-1,
            compute_config=None,
        )
        mock_api_client.start_empty_session_api_v2_sessions_start_empty_session_post.assert_called_once_with(
            session_up_options=SessionUpOptions(
                project_id=project_test_data.id,
                name="session-11",
                cluster_config={
                    "config": json.dumps(yaml.safe_load(CLUSTER_YAML_TEMPLATE))
                },
                cloud_id=cloud_test_data.id,
                idle_timeout=-1,
                start_empty_head_node_only=True,
            )
        )

        mock_api_client.setup_and_initialize_session_api_v2_sessions_session_id_setup_and_initialize_session_post.assert_called_once_with(
            session_id="session-10",
            setup_initialize_session_options=SetupInitializeSessionOptions(
                project_id=project_test_data.id,
                cluster_config={
                    "config": json.dumps(yaml.safe_load(CLUSTER_YAML_TEMPLATE))
                },
                restart_only=False,
                no_restart=False,
                no_update=True,
            ),
        )


def test_anyscale_up_failure(
    mock_up_tuple: Tuple[Mock, Project, Cloud, Mock], mock_auth_api_client
) -> None:
    (mock_api_client, _, _, mock_wait_for_start,) = mock_up_tuple
    session_controller = SessionController()
    session_controller.push = Mock()  # type: ignore
    sesion_id = "session-10"
    mock_api_client.get_project_default_session_name_api_v2_projects_project_id_default_session_name_get.return_value = ProjectdefaultsessionnameResponse(
        result=ProjectDefaultSessionName(name="session-11")
    )
    mock_api_client.list_sessions_api_v2_sessions_get.return_value = SessionListResponse(
        results=[]  # No session with this name yet, brand new session.
    )
    # Session Name should be "11"
    mock_api_client.start_empty_session_api_v2_sessions_start_empty_session_post = Mock(
        return_value=StartemptysessionresponseResponse(
            result=StartEmptySessionResponse(
                session_id=sesion_id, second_update_required=False,
            )
        )
    )
    mock_wait_for_start.side_effect = click.ClickException("Node Failed to Start!")

    with tempfile.NamedTemporaryFile(mode="w") as temp_file:
        temp_file.write(CLUSTER_YAML_TEMPLATE)
        temp_file.flush()
        with pytest.raises(click.ClickException):
            session_controller.up(
                session_name=None,
                config=temp_file.name,
                build_id=None,
                min_workers=None,
                max_workers=None,
                no_restart=False,
                restart_only=False,
                disable_sync=False,
                cloud_id=None,
                cloud_name=None,
                idle_timeout=-1,
                compute_config=None,
            )


def test_anyscale_up_no_config_provided(
    mock_up_tuple: Tuple[Mock, Project, Cloud, Mock], mock_auth_api_client
) -> None:
    (mock_api_client, project_test_data, cloud_test_data, _,) = mock_up_tuple
    session_controller = SessionController()
    session_controller.push = Mock()  # type: ignore
    mock_api_client.get_project_default_session_name_api_v2_projects_project_id_default_session_name_get.return_value = ProjectdefaultsessionnameResponse(
        result=ProjectDefaultSessionName(name="session-11")
    )
    mock_api_client.list_sessions_api_v2_sessions_get.return_value = SessionListResponse(
        results=[
            Session
        ]  # This session already exists so no config needs to be provided
    )
    # Session Name should be "11"
    mock_api_client.start_empty_session_api_v2_sessions_start_empty_session_post = Mock(
        return_value=StartemptysessionresponseResponse(
            result=StartEmptySessionResponse(
                session_id="session-10", second_update_required=False,
            )
        )
    )
    mock_api_client.get_session_cluster_config_api_v2_sessions_session_id_cluster_config_get.return_value = Mock(
        result=ClusterConfig(
            config=CLUSTER_YAML_TEMPLATE, config_with_defaults=CLUSTER_YAML_TEMPLATE
        )
    )
    session_controller.up(
        session_name=None,
        config=None,
        build_id=None,
        min_workers=None,
        max_workers=None,
        no_restart=False,
        restart_only=False,
        disable_sync=False,
        cloud_id=None,
        cloud_name=None,
        idle_timeout=-1,
        compute_config=None,
    )
    mock_api_client.start_empty_session_api_v2_sessions_start_empty_session_post.assert_called_once_with(
        session_up_options=SessionUpOptions(
            project_id=project_test_data.id,
            name="session-11",
            cluster_config={
                "config": json.dumps(yaml.safe_load(CLUSTER_YAML_TEMPLATE))
            },
            cloud_id=cloud_test_data.id,
            idle_timeout=-1,
            start_empty_head_node_only=True,
        )
    )

    mock_api_client.setup_and_initialize_session_api_v2_sessions_session_id_setup_and_initialize_session_post.assert_called_once_with(
        session_id="session-10",
        setup_initialize_session_options=SetupInitializeSessionOptions(
            project_id=project_test_data.id,
            cluster_config={
                "config": json.dumps(yaml.safe_load(CLUSTER_YAML_TEMPLATE))
            },
            restart_only=False,
            no_restart=False,
            no_update=True,
        ),
    )


def test_anyscale_app_template(
    mock_up_tuple: Tuple[Mock, Project, Cloud, Mock], mock_auth_api_client
) -> None:
    (mock_api_client, project_test_data, cloud_test_data, _,) = mock_up_tuple
    session_controller = SessionController()
    session_controller.push = Mock()  # type: ignore
    mock_api_client.get_project_default_session_name_api_v2_projects_project_id_default_session_name_get.return_value = ProjectdefaultsessionnameResponse(
        result=ProjectDefaultSessionName(name="session-11")
    )
    mock_api_client.list_sessions_api_v2_sessions_get.return_value = SessionListResponse(
        results=[]  # No session with this name yet, brand new session.
    )
    # Session Name should be "11"
    mock_api_client.start_empty_session_api_v2_sessions_start_empty_session_post = Mock(
        return_value=StartemptysessionresponseResponse(
            result=StartEmptySessionResponse(
                session_id="session-10", second_update_required=False,
            )
        )
    )
    mock_api_client.create_compute_template_api_v2_compute_templates_post = Mock(
        return_value=Mock(result=Mock(id="mock_compute_template_id"))
    )
    with tempfile.NamedTemporaryFile(mode="w") as temp_file:
        temp_file.write(COMPUTE_CONFIG_TEMPLATE)
        temp_file.flush()

        session_controller.up(
            session_name=None,
            config=None,
            build_id="mock_build_id",
            min_workers=None,
            max_workers=None,
            no_restart=False,
            restart_only=False,
            disable_sync=False,
            cloud_id=None,
            cloud_name=None,
            idle_timeout=-1,
            compute_config=temp_file.name,
        )
        mock_api_client.start_empty_session_api_v2_sessions_start_empty_session_post.assert_called_once_with(
            session_up_options=SessionUpOptions(
                project_id=project_test_data.id,
                name="session-11",
                cluster_config=None,
                build_id="mock_build_id",
                compute_template_id="mock_compute_template_id",
                cloud_id=cloud_test_data.id,
                idle_timeout=-1,
                start_empty_head_node_only=False,
            )
        )

        mock_api_client.setup_and_initialize_session_api_v2_sessions_session_id_setup_and_initialize_session_post.assert_called_once_with(
            session_id="session-10",
            setup_initialize_session_options=SetupInitializeSessionOptions(
                project_id=project_test_data.id,
                cluster_config={"config": "null"},
                restart_only=False,
                no_restart=False,
                no_update=True,
            ),
        )


def test_get_or_generate_session_name(
    mock_up_tuple: Tuple[Mock, Project, Cloud, Mock], mock_auth_api_client
) -> None:
    (mock_api_client, project_test_data, _, _) = mock_up_tuple
    mock_api_client.get_project_default_session_name_api_v2_projects_project_id_default_session_name_get.return_value = ProjectdefaultsessionnameResponse(
        result=ProjectDefaultSessionName(name="session-11")
    )

    session_controller = SessionController()
    session_controller._get_or_generate_session_name(None, project_test_data.id)
    mock_api_client.get_project_default_session_name_api_v2_projects_project_id_default_session_name_get.assert_called_once_with(
        project_id=project_test_data.id
    )


def test_register_compute_config(
    mock_up_tuple: Tuple[Mock, Project, Cloud, Mock], mock_auth_api_client
) -> None:
    (mock_api_client, project_test_data, _, _) = mock_up_tuple
    mock_api_client.create_compute_template_api_v2_compute_templates_post = Mock(
        return_value=Mock(result=Mock(id="mock_compute_template_id"))
    )
    session_controller = SessionController()
    with tempfile.NamedTemporaryFile(mode="w") as temp_file:
        temp_file.write(COMPUTE_CONFIG_TEMPLATE)
        temp_file.flush()
        session_controller._register_compute_config(
            temp_file.name, project_test_data.id
        )

    config_object = ComputeTemplateConfig(**yaml.safe_load(COMPUTE_CONFIG_TEMPLATE))
    create_compute_template = CreateComputeTemplate(
        name=ANY, project_id=project_test_data.id, config=config_object,
    )
    mock_api_client.create_compute_template_api_v2_compute_templates_post.assert_called_once_with(
        create_compute_template=create_compute_template
    )


@pytest.mark.parametrize("update_session", [False, True])
def test_create_or_update_session_data(
    mock_up_tuple: Tuple[Mock, Project, Cloud, Mock],
    mock_auth_api_client,
    update_session: bool,
) -> None:
    (mock_api_client, project_test_data, _, _) = mock_up_tuple
    session_controller = SessionController()
    session_controller.api_client = mock_api_client
    if update_session:
        session_controller.api_client.list_sessions_api_v2_sessions_get.return_value = SessionListResponse(
            results=[Mock(id="session-11-id")]  # Session already exists.
        )
    else:
        session_controller.api_client.list_sessions_api_v2_sessions_get.return_value = SessionListResponse(
            results=[]  # No session with this name yet, brand new session.
        )
    session_controller._create_or_update_session_data(
        session_name="session-11",
        project_id=project_test_data.id,
        build_id="mock_build_id",
        compute_template_id="mock_compute_template_id",
        idle_timeout=-1,
    )

    if update_session:
        session_controller.anyscale_api_client.update_session.assert_called_once_with(
            "session-11-id", UpdateSession(idle_timeout=-1,),
        )
    else:
        create_session_data = CreateSession(
            name="session-11",
            project_id=project_test_data.id,
            cloud_id=None,
            cluster_config=None,
            build_id="mock_build_id",
            compute_template_id="mock_compute_template_id",
            idle_timeout=-1,
            uses_app_config=True,
        )
        session_controller.anyscale_api_client.create_session.assert_called_once_with(
            create_session_data
        )


def test_list_app_configs(
    mock_up_tuple: Tuple[Mock, Project, Cloud, Mock], mock_auth_api_client
) -> None:
    (mock_api_client, project_test_data, _, _) = mock_up_tuple
    session_controller = SessionController()
    session_controller.api_client = mock_api_client
    session_controller.anyscale_api_client.list_app_configs = Mock(
        return_value=Mock(results=[], metadata=Mock(next_paging_token=None))
    )
    session_controller._list_app_configs(project_id=project_test_data.id)
    session_controller.anyscale_api_client.list_app_configs.assert_called_once_with(
        project_id=project_test_data.id, count=50, paging_token=None
    )


def test_list_builds(
    mock_up_tuple: Tuple[Mock, Project, Cloud, Mock], mock_auth_api_client
) -> None:
    (mock_api_client, _, _, _) = mock_up_tuple
    session_controller = SessionController()
    session_controller.api_client = mock_api_client
    session_controller.anyscale_api_client.list_builds = Mock(
        return_value=Mock(results=[], metadata=Mock(next_paging_token=None))
    )
    session_controller._list_builds("mock_app_config_id")
    session_controller.anyscale_api_client.list_builds.assert_called_once_with(
        "mock_app_config_id", count=50, paging_token=None
    )


def test_get_app_config_id_from_name(
    mock_up_tuple: Tuple[Mock, Project, Cloud, Mock], mock_auth_api_client
) -> None:
    (mock_api_client, project_test_data, _, _) = mock_up_tuple
    session_controller = SessionController()
    session_controller.api_client = mock_api_client
    session_controller._list_app_configs = Mock(  # type: ignore
        return_value=[
            AppConfig(
                name="mock_app_config_name",
                id="mock_app_config_id",
                project_id=project_test_data.id,
                organization_id="org_1",
                creator_id=ANY,
                created_at=ANY,
                last_modified_at=ANY,
            )
        ]
    )

    app_config_id = session_controller._get_app_config_id_from_name(
        "mock_app_config_name", project_test_data.id
    )
    assert app_config_id == "mock_app_config_id"
    session_controller._list_app_configs.assert_called_once_with(project_test_data.id)


@pytest.mark.parametrize("build_identifier", ["my_app_config:1", "my_app_config"])
def test_get_build_id_from_build_identifier(
    mock_up_tuple: Tuple[Mock, Project, Cloud, Mock],
    mock_auth_api_client,
    build_identifier: str,
) -> None:
    (mock_api_client, project_test_data, _, _) = mock_up_tuple
    session_controller = SessionController()
    session_controller.api_client = mock_api_client
    session_controller._get_app_config_id_from_name = Mock(  # type: ignore
        return_value="mock_app_config_id"
    )
    session_controller._list_builds = Mock(  # type: ignore
        return_value=[
            Build(
                revision=1,
                id="mock_build_id_1",
                application_template_id="mock_app_config_id",
                config_json=ANY,
                creator_id=ANY,
                status=ANY,
                created_at=ANY,
                last_modified_at=ANY,
            ),
            Build(
                revision=2,
                id="mock_build_id_2",
                application_template_id="mock_app_config_id",
                config_json=ANY,
                creator_id=ANY,
                status=ANY,
                created_at=ANY,
                last_modified_at=ANY,
            ),
        ]
    )
    build_id = session_controller._get_build_id_from_build_identifier(
        build_identifier, project_test_data.id
    )
    if build_identifier == "my_app_config":
        assert build_id == "mock_build_id_2"
    elif build_identifier == "my_app_config:1":
        assert build_id == "mock_build_id_1"

    session_controller._get_app_config_id_from_name.assert_called_once_with(
        "my_app_config", project_test_data.id
    )
    session_controller._list_builds.assert_called_once_with("mock_app_config_id")


@pytest.mark.parametrize("sync_files", [False, True])
@pytest.mark.parametrize("use_build_id", [False, True])
def test_anyscale_start(
    mock_up_tuple: Tuple[Mock, Project, Cloud, Mock],
    mock_auth_api_client,
    sync_files: bool,
    use_build_id: bool,
) -> None:
    (mock_api_client, project_test_data, _, _,) = mock_up_tuple
    session_controller = SessionController()
    session_controller.api_client = mock_api_client
    session_controller.push = Mock()  # type: ignore
    session_controller._get_or_generate_session_name = Mock(return_value="session-11")  # type: ignore
    session_controller._register_compute_config = Mock(  # type: ignore
        return_value="mock_compute_template_id"
    )
    session_controller._create_or_update_session_data = Mock(  # type: ignore
        return_value=Mock(id="session-11-id")
    )
    session_controller._get_build_id_from_build_identifier = Mock(  # type: ignore
        return_value="mock_build_id"
    )

    if sync_files:
        # Check --sync-files is deprecated
        with pytest.raises(click.ClickException):
            session_controller.start(
                session_name=None,
                build_id=None,
                build_identifier="mock_build_identifier",
                compute_config="mock_compute_config_file",
                idle_timeout=-1,
                sync_files=sync_files,
            )
    elif use_build_id:
        session_controller.start(
            session_name=None,
            build_id="mock_build_id",
            build_identifier=None,
            compute_config="mock_compute_config_file",
            idle_timeout=-1,
            sync_files=sync_files,
        )
    else:
        session_controller.start(
            session_name=None,
            build_id=None,
            build_identifier="mock_build_identifier",
            compute_config="mock_compute_config_file",
            idle_timeout=-1,
            sync_files=sync_files,
        )

    if not sync_files:
        session_controller._get_or_generate_session_name.assert_called_once_with(
            None, project_test_data.id
        )
        session_controller._register_compute_config.assert_called_once_with(
            "mock_compute_config_file", project_test_data.id
        )
        session_controller._create_or_update_session_data.assert_called_once_with(
            "session-11",
            project_test_data.id,
            "mock_build_id",
            "mock_compute_template_id",
            -1,
        )
        session_controller.anyscale_api_client.start_session.assert_called_once_with(
            "session-11-id",
            StartSessionOptions(
                build_id="mock_build_id", compute_template_id="mock_compute_template_id"
            ),
        )

        if not use_build_id:
            session_controller._get_build_id_from_build_identifier.assert_called_once_with(
                "mock_build_identifier", project_test_data.id
            )


@pytest.mark.parametrize("force_head_node", [True, False])
@pytest.mark.parametrize("on_head_node", [True, False])
@pytest.mark.parametrize("worker_node_id", [None, "123456"])
@pytest.mark.parametrize("worker_node_ip", [None, "0.0.0.0"])
def test_ssh_session(
    session_test_data: Session,
    mock_auth_api_client_multiple_sessions,
    on_head_node: bool,
    worker_node_id: Optional[str],
    worker_node_ip: Optional[str],
    force_head_node: bool,
) -> None:
    session_controller = SessionController()

    mock_project_definition = Mock()
    mock_project_definition.root = "/some/directory"
    mock_load_project_or_throw = Mock(return_value=mock_project_definition)

    mock_get_project_id = Mock(return_value=session_test_data.project_id)
    mock_get_project_session = Mock(return_value=session_test_data)
    mock_cluster_config = {
        "auth": {"ssh_user": "ubuntu", "ssh_private_key": ""},
        "docker": {"container_name": "ray_container"},
    }
    mock_get_cluster_config = Mock(return_value=mock_cluster_config)
    mock_ray_init = Mock()
    mock_node_manager_address = "111.111.111.111"
    mock_node_data = [
        {"NodeID": "123456", "NodeManagerAddress": mock_node_manager_address}
    ]
    mock_ray_nodes = Mock(return_value=mock_node_data)

    mock_exists = Mock(return_value=on_head_node)
    mock_get_and_validate_project_id = Mock()

    with patch.multiple(
        "anyscale.controllers.session_controller",
        load_project_or_throw=mock_load_project_or_throw,
        get_project_id=mock_get_project_id,
        get_cluster_config=mock_get_cluster_config,
        get_project_session=mock_get_project_session,
        get_and_validate_project_id=mock_get_and_validate_project_id,
    ), patch.multiple("subprocess", run=Mock()), patch(
        "builtins.open", mock_open(read_data=str(mock_cluster_config))
    ), patch.multiple(
        "ray", init=mock_ray_init, nodes=mock_ray_nodes
    ), patch.multiple(
        "os.path", exists=mock_exists
    ):
        if force_head_node and not on_head_node:
            # It's the user's responsibility to never do this
            return
        if (force_head_node and worker_node_id is None and worker_node_ip is None) or (
            worker_node_id is not None and worker_node_ip is not None
        ):
            with pytest.raises(click.ClickException):
                session_controller.ssh(
                    session_name=session_test_data.name,
                    ssh_option=("",),
                    worker_node_id=worker_node_id,
                    worker_node_ip=worker_node_ip,
                    force_head_node=force_head_node,
                    project_id=None,
                )
            return
        else:
            session_controller.ssh(
                session_name=session_test_data.name,
                ssh_option=("",),
                worker_node_id=worker_node_id,
                worker_node_ip=worker_node_ip,
                force_head_node=force_head_node,
                project_id=None,
            )

    if on_head_node and worker_node_id is not None:
        mock_ray_init.assert_called_once_with(address="auto")
        mock_ray_nodes.assert_called_once_with()
    elif on_head_node and worker_node_ip is not None:
        mock_ray_init.assert_not_called()
        mock_ray_nodes.assert_not_called()
    elif worker_node_id is not None or worker_node_ip is not None:
        session_controller.api_client.get_session_head_ip_api_v2_sessions_session_id_head_ip_get.assert_called_once_with(
            session_test_data.id,
        )
    else:
        session_controller.api_client.get_session_head_ip_api_v2_sessions_session_id_head_ip_get.assert_called_once_with(
            session_test_data.id,
        )
        mock_get_and_validate_project_id.assert_called_once_with(
            project_id=None,
            project_name=None,
            api_client=session_controller.api_client,
            anyscale_api_client=session_controller.anyscale_api_client,
        )


def test_anyscale_push_session(
    session_test_data: Session, mock_auth_api_client_with_session
) -> None:
    mock_project_definition = Mock()
    mock_project_definition.root = "/some/directory"
    mock_load_project_or_throw = Mock(return_value=mock_project_definition)
    mock_get_project_id = Mock(return_value=session_test_data.project_id)
    mock_get_project_session = Mock(return_value=session_test_data)
    cluster_config_data = {"auth": {"ssh_user": "", "ssh_private_key": ""}}
    mock_get_cluster_config = Mock(return_value=cluster_config_data)
    mock_subprocess_run = Mock()
    mock_subprocess_run.return_value.returncode = 0

    mock_validate_cluster_configuration = Mock()
    mock_rsync = Mock()
    mock_up = Mock()

    session_controller = SessionController()

    with patch.multiple("subprocess", run=mock_subprocess_run), patch.multiple(
        "anyscale.controllers.session_controller",
        load_project_or_throw=mock_load_project_or_throw,
        get_project_id=mock_get_project_id,
        get_cluster_config=mock_get_cluster_config,
        get_project_session=mock_get_project_session,
        validate_cluster_configuration=mock_validate_cluster_configuration,
        rsync=mock_rsync,
    ), patch.object(session_controller, "up", new=mock_up):
        session_name = "TestSessionName"
        source = "TestSource"
        target = "TestTarget"
        config = "TestConfig"
        all_nodes = False
        session_controller.push(session_name, source, target, config, all_nodes)

        # Don't check the config_file argument because it's a random path of
        # a temporary file
        mock_rsync.assert_called_once_with(
            ANY, source=source, target=target, down=False,
        )

        mock_validate_cluster_configuration.assert_called_once_with(
            config, api_instance=session_controller.api_client
        )

        mock_up.assert_called_once_with(
            session_name=session_test_data.name,
            config=None,
            build_id=None,
            compute_config=None,
            min_workers=None,
            max_workers=None,
            no_restart=False,
            idle_timeout=None,
            restart_only=False,
            disable_sync=True,
            cloud_id=session_test_data.cloud_id,
            cloud_name=None,
        )


@pytest.mark.parametrize("config", [None, "tmp.yaml"])
def test_pull(
    session_test_data: Session, mock_auth_api_client_with_session, config: str,
) -> None:
    mock_project_definition = Mock()
    mock_project_definition.root = "/some/directory"
    mock_load_project_or_throw = Mock(return_value=mock_project_definition)
    mock_get_project_id = Mock(return_value=session_test_data.project_id)
    cluster_config_data = {"auth": {"ssh_user": "", "ssh_private_key": ""}}
    mock_get_cluster_config = Mock(return_value=cluster_config_data)
    mock_get_working_dir = Mock(return_value="directory")
    mock_rsync = Mock()
    mock_get_project_session = Mock(return_value=session_test_data)

    session_controller = SessionController()
    session_controller.api_client.get_session_cluster_config_api_v2_sessions_session_id_cluster_config_get.return_value.result.config_with_defaults = (
        "temp"
    )

    with patch.multiple(
        "anyscale.controllers.session_controller",
        load_project_or_throw=mock_load_project_or_throw,
        get_project_id=mock_get_project_id,
        get_cluster_config=mock_get_cluster_config,
        get_working_dir=mock_get_working_dir,
        get_project_session=mock_get_project_session,
        rsync=mock_rsync,
    ), patch("builtins.open", new_callable=mock_open()), patch("yaml.dump"):
        session_controller.pull(session_name=session_test_data.name, config=config)

    mock_load_project_or_throw.assert_called_once_with()
    mock_get_project_id.assert_called_once_with("/some/directory")
    mock_get_cluster_config.assert_called_once_with(
        session_test_data.name, session_controller.api_client
    )
    mock_get_working_dir.assert_called_once_with(
        cluster_config_data, session_test_data.project_id, session_controller.api_client
    )
    mock_rsync.assert_called_once()
    if config:
        mock_get_project_session.assert_called_once_with(
            session_test_data.project_id,
            session_test_data.name,
            session_controller.api_client,
        )
        session_controller.api_client.get_session_cluster_config_api_v2_sessions_session_id_cluster_config_get.assert_called_once_with(
            session_test_data.id
        )


@pytest.mark.parametrize("project_name", [None, "project_name"])
def test_resolve_session(
    project_test_data: Project,
    session_test_data: Session,
    mock_auth_api_client_with_session,
    project_name: Optional[str],
) -> None:
    mock_project_definition = Mock()
    mock_project_definition.root = "/some/directory"
    mock_load_project_or_throw = Mock(return_value=mock_project_definition)

    mock_get_project_id = Mock(return_value=session_test_data.project_id)

    session_controller = SessionController()
    session_controller.api_client.list_sessions_api_v2_sessions_get.return_value = SessionListResponse(
        results=[session_test_data]
    )
    session_controller.api_client.list_projects_api_v2_projects_get.return_value = ProjectListResponse(
        results=[project_test_data]
    )

    with patch.multiple(
        "anyscale.controllers.session_controller",
        load_project_or_throw=mock_load_project_or_throw,
        get_project_id=mock_get_project_id,
    ):
        session = session_controller._resolve_session("session_name", project_name)

    assert session == session_test_data

    if project_name:
        mock_load_project_or_throw.assert_not_called()
        mock_get_project_id.assert_not_called()
        session_controller.api_client.list_projects_api_v2_projects_get.assert_called_once_with()
    else:
        mock_load_project_or_throw.assert_called_once()
        mock_get_project_id.assert_called_once_with(mock_project_definition.root)
        session_controller.api_client.list_projects_api_v2_projects_get.assert_not_called()

    session_controller.api_client.list_sessions_api_v2_sessions_get.assert_called_once_with(
        session_test_data.project_id, name="session_name"
    )



from contextlib import contextmanager
from copy import deepcopy
import datetime
import json
import logging
import os
import random
import string
import subprocess
import sys
import threading
import time
from typing import Any, Callable, cast, Dict, Generator, Iterator, List, Optional, Tuple
from urllib.parse import quote as quote_sanitize, urlencode, urljoin
import webbrowser

from botocore.exceptions import NoRegionError
import click
import requests
from requests import Response
import yaml

from anyscale.authenticate import get_auth_api_client
from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.api.default_api import DefaultApi
from anyscale.client.openapi_client.models.user_info import UserInfo
from anyscale.client.openapi_client.rest import ApiException
import anyscale.conf
import anyscale.shared_anyscale_utils.conf as shared_anyscale_conf
from anyscale.shared_anyscale_utils.util import get_container_name


logger = logging.getLogger(__file__)

BOTO_MAX_RETRIES = 5
PROJECT_NAME_ENV_VAR = "ANYSCALE_PROJECT_NAME"

VALID_PYTHON_VERSIONS = ["py36", "py37", "py38", "py39"]

log = BlockLogger()  # Anyscale CLI Logger

# Cached 07-21-2021
_CACHED_GCP_REGIONS = json.loads(
    open(os.path.join(os.path.dirname(__file__), "cached_gcp_regions.json")).read()
)["regions"]


def confirm(msg: str, yes: bool) -> Any:
    return None if yes else click.confirm(msg, abort=True)


class AnyscaleEndpointFormatter:
    def __init__(self, host: Optional[str] = None):
        self.host = host or shared_anyscale_conf.ANYSCALE_HOST

    def get_endpoint(self, endpoint: str) -> str:
        return str(urljoin(self.host, endpoint))

    def get_job_endpoint(self, job_id: str) -> str:
        return self.get_endpoint(f"/jobs/{job_id}")


def get_endpoint(endpoint: str, host: Optional[str] = None) -> str:
    return str(urljoin(host or shared_anyscale_conf.ANYSCALE_HOST, endpoint))


def send_json_request_raw(
    endpoint: str,
    json_args: Dict[str, Any],
    method: str = "GET",
    cli_token: Optional[str] = None,
    host: Optional[str] = None,
) -> Response:
    get_auth_api_client(cli_token=cli_token, host=host)

    url = get_endpoint(endpoint, host=host)
    cookies = {"cli_token": cli_token or anyscale.conf.CLI_TOKEN or ""}
    try:
        if method == "GET":
            resp = requests.get(url, params=json_args, cookies=cookies)
        elif method == "POST":
            resp = requests.post(url, json=json_args, cookies=cookies)
        elif method == "DELETE":
            resp = requests.delete(url, json=json_args, cookies=cookies)
        elif method == "PATCH":
            resp = requests.patch(url, data=json_args, cookies=cookies)
        elif method == "PUT":
            resp = requests.put(url, json=json_args, cookies=cookies)
        else:
            assert False, "unknown method {}".format(method)
    except requests.exceptions.ConnectionError:
        raise click.ClickException(
            "Failed to connect to anyscale server at {}".format(url)
        )

    return resp


def send_json_request(
    endpoint: str,
    json_args: Dict[str, Any],
    method: str = "GET",
    cli_token: Optional[str] = None,
    host: Optional[str] = None,
) -> Dict[str, Any]:
    resp = send_json_request_raw(
        endpoint, json_args, method=method, cli_token=cli_token, host=host,
    )

    if not resp.ok:
        if resp.status_code == 500:
            raise click.ClickException(
                "There was an internal error in this command. "
                "Please report this to the Anyscale team at support@anyscale.zendesk.com "
                "with the token '{}'.".format(resp.headers["x-trace-id"])
            )

        raise click.ClickException("{}: {}.".format(resp.status_code, resp.text))

    if resp.status_code == 204:
        return {}

    json_resp: Dict[str, Any] = resp.json()
    if "error" in json_resp:
        raise click.ClickException("{}".format(json_resp["error"]))

    return json_resp


def deserialize_datetime(s: str) -> datetime.datetime:
    if sys.version_info < (3, 7) and ":" == s[-3:-2]:
        s = s[:-3] + s[-2:]

    return datetime.datetime.strptime(s, "%Y-%m-%dT%H:%M:%S.%f%z")


def humanize_timestamp(timestamp: datetime.datetime) -> str:
    delta = datetime.datetime.now(datetime.timezone.utc) - timestamp
    offset = float(delta.seconds + (delta.days * 60 * 60 * 24))
    delta_s = int(offset % 60)
    offset /= 60
    delta_m = int(offset % 60)
    offset /= 60
    delta_h = int(offset % 24)
    offset /= 24
    delta_d = int(offset)

    if delta_d >= 1:
        return "{} day{} ago".format(delta_d, "s" if delta_d > 1 else "")
    if delta_h > 0:
        return "{} hour{} ago".format(delta_h, "s" if delta_h > 1 else "")
    if delta_m > 0:
        return "{} minute{} ago".format(delta_m, "s" if delta_m > 1 else "")
    else:
        return "{} second{} ago".format(delta_s, "s" if delta_s > 1 else "")


def get_cluster_config(config_path: str) -> Any:
    with open(config_path) as f:
        cluster_config = yaml.safe_load(f)

    return cluster_config


def get_requirements(requirements_path: str) -> str:
    with open(requirements_path) as f:
        return f.read()


def _resource(name: str, region: str) -> Any:
    import boto3
    from botocore.config import Config

    boto_config = Config(retries={"max_attempts": BOTO_MAX_RETRIES})
    return boto3.resource(name, region, config=boto_config)


def _client(name: str, region: str) -> Any:
    return _resource(name, region).meta.client


def _get_role(role_name: str, region: str) -> Any:
    import botocore

    iam = _resource("iam", region)
    role = iam.Role(role_name)
    try:
        role.load()
        return role
    except botocore.exceptions.ClientError as exc:
        if exc.response.get("Error", {}).get("Code") == "NoSuchEntity":
            return None
        else:
            raise exc


def _get_user(user_name: str, region: str) -> Any:
    import botocore

    iam = _resource("iam", region)
    user = iam.User(user_name)
    try:
        user.load()
        return user
    except botocore.exceptions.ClientError as exc:
        if exc.response.get("Error", {}).get("Code") == "NoSuchEntity":
            return None
        else:
            raise exc


def get_available_regions() -> List[str]:
    import boto3

    try:
        client = boto3.client("ec2")
    except NoRegionError:
        # If there is no region, default to `us-west-2`
        client = boto3.client("ec2", region_name="us-west-2")
    return [region["RegionName"] for region in client.describe_regions()["Regions"]]


def launch_gcp_cloud_setup(
    name: str,
    region: str,
    folder_id: Optional[int],
    vpc_peering_ip_range: Optional[str],
    vpc_peering_target_project_id: Optional[str],
    vpc_peering_target_vpc_id: Optional[str],
) -> None:
    # TODO: enforce uniqueness for user's clouds
    quote_safe_name = quote_sanitize(name, safe="")
    query_params = {
        "region": region,
    }
    if folder_id:
        query_params["folder_id"] = str(folder_id)
    if (
        vpc_peering_ip_range
        and vpc_peering_target_project_id
        and vpc_peering_target_vpc_id
    ):
        query_params.update(
            {
                "vpc_peering_ip_range": vpc_peering_ip_range,
                "vpc_peering_target_project_id": vpc_peering_target_project_id,
                "vpc_peering_target_vpc_id": vpc_peering_target_vpc_id,
            }
        )
    # TODO: Replace this with a proper endpoint
    endpoint = f"/api/v2/clouds/gcp/create/{quote_safe_name}?{urlencode(query_params)}"
    full_url = get_endpoint(endpoint)
    print(
        f"Launching GCP Oauth Flow:\n{full_url}\n(If this window does not auto-launch, use the link above)"
    )
    webbrowser.open(full_url)


class Timer:
    """
    Code adopted from https://stackoverflow.com/a/39504463/3727678
    Spawn thread and time process that may be blocking.
    """

    def timer_generator(self) -> Iterator[str]:
        while True:
            time_diff = time.gmtime(time.time() - self.start_time)
            yield "{0}: {1}".format(self.message, time.strftime("%M:%S", time_diff))

    def __init__(self, message: str = "") -> None:
        self.message = message
        self.busy = False
        self.start_time = 0.0

    def timer_task(self) -> None:
        while self.busy:
            sys.stdout.write(next(self.timer_generator()))
            sys.stdout.flush()
            time.sleep(0.1)
            sys.stdout.write("\b" * (len(self.message) + 20))
            sys.stdout.flush()

    def start(self) -> None:
        self.busy = True
        self.start_time = time.time()
        threading.Thread(target=self.timer_task).start()

    def stop(self) -> None:
        sys.stdout.write("\n")
        sys.stdout.flush()
        self.busy = False
        self.start_time = 0.0
        time.sleep(1)


def check_is_feature_flag_on(flag_key: str, default: bool = False) -> bool:
    try:
        is_on = send_json_request(
            "/api/v2/userinfo/check_is_feature_flag_on", {"flag_key": flag_key},
        )
    except Exception:
        return default

    return cast(bool, is_on["result"]["is_on"])


def get_active_sessions(
    project_id: str, session_name: str, api_client: Optional[DefaultApi]
) -> Any:
    if api_client:
        return api_client.list_sessions_api_v2_sessions_get(
            project_id=project_id, name=session_name, active_only=True
        ).results
    else:
        response = anyscale.util.send_json_request(
            "/api/v2/sessions/",
            {"project_id": project_id, "name": session_name, "active_only": True},
        )
        return response["results"]


def get_project_directory_name(project_id: str, api_client: DefaultApi = None) -> str:
    if api_client is None:
        api_client = get_auth_api_client().api_client

    # TODO (yiran): return error early if project doesn't exist.
    resp = api_client.get_project_api_v2_projects_project_id_get(project_id)
    directory_name = resp.result.directory_name
    assert len(directory_name) > 0, "Empty directory name found."
    return cast(str, directory_name)


def get_working_dir(
    cluster_config: Dict[str, Any], project_id: str, api_client: DefaultApi = None
) -> str:
    working_dir: Optional[str] = (
        cluster_config.get("metadata", {}).get("anyscale", {}).get("working_dir")
    )
    if working_dir:
        return working_dir
    else:
        return f"/home/ray/{get_project_directory_name(project_id, api_client)}"


def get_wheel_url(
    ray_commit: str,
    ray_version: str,
    py_version: Optional[str] = None,
    sys_platform: Optional[str] = None,
) -> str:
    """Return S3 URL for the given release spec or 'latest'."""
    if py_version is None:
        py_version = "".join(str(x) for x in sys.version_info[0:2])
    if sys_platform is None:
        sys_platform = sys.platform

    if sys_platform == "darwin":
        if py_version == "38":
            platform = "macosx_10_15_x86_64"
        else:
            platform = "macosx_10_15_intel"
    elif sys_platform == "win32":
        platform = "win_amd64"
    else:
        platform = "manylinux2014_x86_64"

    if py_version == "38":
        py_version_malloc = py_version
    else:
        py_version_malloc = f"{py_version}m"

    if "dev" in ray_version:
        ray_release = f"master/{ray_commit}"
    else:
        ray_release = f"releases/{ray_version}/{ray_commit}"
    return (
        "https://s3-us-west-2.amazonaws.com/ray-wheels/"
        "{}/ray-{}-cp{}-cp{}-{}.whl".format(
            ray_release, ray_version, py_version, py_version_malloc, platform
        )
    )


@contextmanager
def updating_printer() -> Generator[Callable[[str], None], None, None]:
    import shutil

    cols, _ = shutil.get_terminal_size()

    def print_status(status: str) -> None:
        lines = status.splitlines()
        first_line = lines[0]
        truncated_first_line = (
            first_line[0:cols]
            if len(first_line) <= cols and len(lines) == 1
            else (first_line[0 : cols - 3] + "...")
        )
        # Clear the line first
        print("\r" + " " * cols, end="\r")
        print(truncated_first_line, end="", flush=True)

    try:
        yield print_status
    finally:
        # Clear out the status and return to the beginning to reprint
        print("\r" + " " * cols, end="\r", flush=True)


def wait_for_session_start(
    project_id: str,
    session_name: str,
    api_client: Optional[DefaultApi] = None,
    log: BlockLogger = log,
    block_label: Optional[str] = None,
) -> str:
    if block_label:
        log.info(
            f"Waiting for cluster {BlockLogger.highlight(session_name)} to start. This may take a few minutes",
            block_label=block_label,
        )
    else:
        log.info(
            f"Waiting for cluster {session_name} to start. This may take a few minutes"
        )

    if api_client is None:
        api_client = get_auth_api_client().api_client

    with updating_printer() as print_status:
        while True:
            sessions = api_client.list_sessions_api_v2_sessions_get(
                project_id=project_id, name=session_name, active_only=False
            ).results

            if len(sessions) > 0:
                session = sessions[0]

                if session.state == "Running" and session.pending_state is None:
                    return cast(str, session.id)

                # Check for start up errors
                if (
                    session.state_data
                    and session.state_data.startup
                    and session.state_data.startup.startup_error
                ):
                    raise click.ClickException(
                        f"Error while starting cluster {session_name}: {session.state_data.startup.startup_error}"
                    )
                elif (
                    session.state
                    and "Errored" in session.state
                    and session.pending_state is None
                ):
                    raise click.ClickException(
                        f"Error while starting cluster {session_name}: Cluster startup failed due to an error ({session.state})."
                    )
                elif (
                    session.state
                    and session.state in {"Terminated", "Stopped"}
                    and session.pending_state is None
                ):
                    # Cluster is created in Terminated state; Check pending state to see if it is pending transition.
                    raise click.ClickException(
                        f"Error while starting cluster {session_name}: Cluster is still in stopped/terminated state."
                    )
                elif (
                    session.state_data
                    and session.state_data.startup
                    and session.state_data.startup.startup_progress
                ):
                    # Print the latest status
                    print_status(
                        "Starting up " + session.state_data.startup.startup_progress
                    )
                elif (
                    session.state != "StartingUp"
                    and session.pending_state == "StartingUp"
                ):
                    print_status("Waiting for start up...")
            else:
                raise click.ClickException(
                    f"Error while starting cluster {session_name}: Cluster doesn't exist."
                )

            time.sleep(2)


def wait_for_head_node_start(
    project_id: str,
    session_name: str,
    session_id: str,
    api_client: Optional[DefaultApi] = None,
) -> str:
    url = get_endpoint(f"/projects/{project_id}/clusters/{session_id}")
    log.info(
        f"Waiting for head node of cluster {session_name} to start. "
        f"Please view the startup logs at {url}"
    )
    if api_client is None:
        api_client = get_auth_api_client().api_client

    retry_count = 0
    while True:
        sessions = api_client.list_sessions_api_v2_sessions_get(
            project_id=project_id, name=session_name, active_only=False
        ).results

        if len(sessions) > 0:
            session = sessions[0]

            if (
                session.state == "Running" and session.pending_state is None
            ) or session.state == "AwaitingFileMounts":
                return cast(str, session.id)

            # Check for start up errors
            if (
                session.state_data
                and session.state_data.startup
                and session.state_data.startup.startup_error
            ):
                raise click.ClickException(
                    f"Error while starting cluster {session_name}: {session.state_data.startup.startup_error}"
                )
            elif session.state and "Errored" in session.state:
                raise click.ClickException(
                    f"Error while starting cluster {session_name}: Cluster startup failed due to an error ({session.state})."
                )
            elif (
                retry_count >= 2
                and session.state
                and session.state in {"Terminated", "Stopped"}
                and session.pending_state is None
            ):
                # Cluster is created in Terminated state; Ignore the first few state checks.
                raise click.ClickException(
                    f"Error while starting cluster {session_name}: Cluster is still in stopped/terminated state."
                )
        else:
            raise click.ClickException(
                f"Error while starting cluster {session_name}: Cluster doesn't exist."
            )

        time.sleep(5)
        retry_count += 1


def validate_cluster_configuration(
    cluster_config_file_name: str,
    cluster_config: Optional[DefaultApi] = None,
    api_instance: Optional[DefaultApi] = None,
) -> None:
    assert api_instance

    if not os.path.isfile(cluster_config_file_name):
        raise click.ClickException(
            "The configuration file {} does not exist. Please provide a valid config file.".format(
                cluster_config_file_name
            )
        )

    if not cluster_config:
        try:
            with open(cluster_config_file_name) as f:
                cluster_config = yaml.safe_load(f)
        except (ValueError, yaml.YAMLError):
            raise click.ClickException(
                "\tThe configuration file {} does not have a valid format. "
                "\n\tPlease look at https://github.com/ray-project/ray/blob/master/python/ray/autoscaler/aws/example-full.yaml "
                "for an example configuration file.".format(cluster_config_file_name)
            )

    try:
        api_instance.validate_cluster_api_v2_sessions_validate_cluster_post(
            body={"config": json.dumps(cluster_config)}
        )
    except ApiException as e:
        error = json.loads(json.loads(e.body)["error"]["detail"])
        path = ".".join(error["path"])
        if error["path"]:
            formatted_error = 'Error occured at "{k}: {v}" because {message}.\nSchema description for {k}:\n{schema}'.format(
                k=path,
                v=error["instance"],
                message=error["message"],
                schema=json.dumps(error["schema"], indent=4, sort_keys=True),
            )
        else:
            formatted_error = 'Error occured at "{v}" because {message}.\nSchema description:\n{schema}'.format(
                v=error["instance"],
                message=error["message"],
                schema=json.dumps(error["schema"], indent=4, sort_keys=True),
            )
        raise click.ClickException(
            "The configuration file {0} is not valid.\n{1}".format(
                cluster_config_file_name, formatted_error
            )
        )


def _get_rsync_args(rsync_exclude: List[str], rsync_filter: List[str]) -> List[str]:
    rsync_exclude_args = [["--exclude", exclude] for exclude in rsync_exclude]
    rsync_filter_args = [["--filter", f"dir-merge,- {f}"] for f in rsync_filter]

    # Combine and flatten the two lists
    return [
        arg for sublist in rsync_exclude_args + rsync_filter_args for arg in sublist
    ]


def get_rsync_command(
    ssh_command: List[str],
    source: str,
    ssh_user: str,
    head_ip: str,
    target: str,
    delete: bool,
    rsync_exclude: Optional[List[str]] = None,
    rsync_filter: Optional[List[str]] = None,
    dry_run: bool = False,
) -> Tuple[List[str], Optional[Dict[str, Any]]]:
    rsync_executable = "rsync"
    env = None
    if rsync_exclude is None:
        rsync_exclude = []
    if rsync_filter is None:
        rsync_filter = []

    command_list = [
        rsync_executable,
        "--rsh",
        " ".join(ssh_command),
        "-avz",
    ]
    command_list += _get_rsync_args(rsync_exclude, rsync_filter)

    if delete:
        # Deletes files in target that doesn't exist in source
        command_list.append("--delete")

    if dry_run:
        command_list += [
            "--dry-run",
            "--itemize-changes",
            "--out-format",
            "%o %f",
        ]

    command_list += [
        source,
        "{}@{}:{}".format(ssh_user, head_ip, target),
    ]
    return command_list, env


def populate_session_args(cluster_config_str: str, config_file_name: str) -> str:
    import jinja2

    env = jinja2.Environment()
    t = env.parse(cluster_config_str)
    for elem in t.body[0].nodes:  # type: ignore

        if isinstance(elem, jinja2.nodes.Getattr):  # type: ignore
            if elem.attr not in os.environ:
                prefixed_command = " ".join(
                    [f"{elem.attr}=<value>", "anyscale"] + sys.argv[1:]
                )
                raise click.ClickException(
                    f"\tThe environment variable {elem.attr} was not set, yet it is required "
                    f"for configuration file {config_file_name}.\n\tPlease specify {elem.attr} "
                    f"by prefixing the command.\n\t\t{prefixed_command}"
                )

    template = jinja2.Template(cluster_config_str)
    cluster_config_filled = template.render(env=os.environ)
    return cluster_config_filled


def canonicalize_remote_location(
    cluster_config: Dict[str, Any], remote_location: Optional[str], project_id: str
) -> Optional[str]:
    """Returns remote_location, but changes it from being based
    in "~/", "/root/" or "/ray" to match working_dir
    """
    # Include the /root path to ensure that absolute paths also work
    # This is because of an implementation detail in OSS Ray's rsync
    if bool(get_container_name(cluster_config)) and bool(remote_location):
        remote_location = str(remote_location)
        working_dir = get_working_dir(cluster_config, project_id)
        for possible_name in ["root", "ray"]:
            full_name = f"/{possible_name}/"
            # TODO(ilr) upstream this to OSS Ray
            if working_dir.startswith("~/") and remote_location.startswith(full_name):
                return remote_location.replace(full_name, "~/", 1)

            if working_dir.startswith(full_name) and remote_location.startswith("~/"):
                return remote_location.replace("~/", full_name, 1)

    return remote_location


def get_user_info() -> Optional[UserInfo]:
    try:
        api_client = get_auth_api_client().api_client
    except click.exceptions.ClickException:
        return None
    return api_client.get_user_info_api_v2_userinfo_get().result


def get_anyscale_version() -> str:
    try:
        # Return git sha if anyscale pip package was built in development mode
        # TODO (aguo): Pip version 21.2.1 made a change to the FrozenRequirement api. Get rid
        # of this interal api usage.
        from pip._internal.operations.freeze import FrozenRequirement
        import pkg_resources

        distributions = {v.key: v for v in pkg_resources.working_set}
        distribution = distributions["anyscale"]
        frozen_requirement = FrozenRequirement.from_dist(distribution)  # type: ignore
        if frozen_requirement.editable:
            try:
                return (
                    subprocess.check_output(  # noqa: B1
                        ["git", "describe", "--always"],
                        cwd=os.path.dirname(os.path.realpath(__file__)),
                    )  # noqa: B1
                    .strip()
                    .decode("utf-8")
                )
            except subprocess.CalledProcessError:
                log.warning(
                    "Not a git repository, will use standard release versioning"
                )
    except Exception:
        # None of this is critical behavior and so this should not crash the CLI if we can't get a git version.
        pass
    return anyscale.__version__


def generate_slug(length: int = 6) -> str:
    return "".join(random.choices(string.ascii_lowercase + string.digits, k=length))


def validate_non_negative_arg(ctx, param, value):
    """
    Checks that an integer option to click command is non-negative.
    """
    if value < 0:
        raise click.ClickException(
            f"Please specify a non-negative value for {param.opts[0]}"
        )
    return value


def _update_external_ids_for_policy(
    original_policy: Dict[str, Any], new_external_id: str
):
    """Gets All External IDs From policy Dict."""
    policy = deepcopy(original_policy)
    external_ids = [
        statement.setdefault("Condition", {})
        .setdefault("StringEquals", {})
        .setdefault("sts:ExternalId", [])
        for statement in policy.get("Statement", [])
    ]

    external_ids = [
        [i, new_external_id] if isinstance(i, str) else i + [new_external_id]
        for i in external_ids
    ]

    _ = [
        policy["Statement"][i]["Condition"]["StringEquals"].update(
            {"sts:ExternalId": external_ids[i]}
        )
        for i in range(len(policy.get("Statement", [])))
    ]
    return policy


def extract_versions_from_image_name(image_name: str) -> Tuple[str, str]:
    """Returns the python version and ray extracted from an image tag.

    Args:
        image_name: e.g. anyscale/ray-ml:1.11.1-py38-gpu

    Returns:
        The (python version, ray version), e.g. ("py38", "1.11.1")
    """
    # e.g. 1.11.1-py32-gpu
    image_version = image_name.split(":")[-1]
    parts = image_version.split("-")
    # e.g. 1.11.1

    if len(parts) < 2:
        raise ValueError(
            f"Expected the docker image name have an image version tag (something like ray-ml:1.11.1-py38-gpu), got {image_version}."
        )

    ray_version = parts[0]
    # Verify ray_version is valid.
    _ray_version_major_minor(ray_version)

    python_version = parts[1]
    _check_python_version(python_version)
    return (python_version, ray_version)


def _ray_version_major_minor(ray_version: str) -> Tuple[int, int]:
    """Takes in a Ray version such as "1.9.0rc1", "1.10.2", "2.0.0dev0".

    Returns the major minor pair e.g. (1,9) (1,10) (2,0).

    To avoid introducing undesirable dependencies, partly duplicates logic from the Anyscale
    backend.
    """
    invalid_ray_version_msg = (
        f"The Ray version `{ray_version}` has an unexpected format."
    )
    version_components = ray_version.split(".")
    assert len(version_components) >= 2, invalid_ray_version_msg
    major_str, minor_str = version_components[:2]
    assert major_str.isnumeric() and minor_str.isnumeric(), invalid_ray_version_msg
    major_int, minor_int = int(major_str), int(minor_str)
    return (major_int, minor_int)


def _check_python_version(python_version: str) -> None:
    assert (
        python_version in VALID_PYTHON_VERSIONS
    ), f"Expected python_version to be one of {VALID_PYTHON_VERSIONS}, got {python_version}."


def sleep_till(wake_time: float) -> None:
    """ Sleep till the designated time
    """
    time.sleep(max(0, wake_time - time.time()))


def poll(
    interval_secs: float = 1,
    timeout_secs: Optional[float] = None,
    max_iter: Optional[int] = None,
) -> Generator[int, None, None]:
    """ Poll every interval_secs, until timeout_secs, or max iterations has been reached.
    Yield the iteration number, starting at 1.
    """
    last_poll_time = time.time()
    end_time = time.time() + timeout_secs if timeout_secs else None
    count = 0
    should_continue_iter = max_iter is None or count < max_iter
    should_continue_time = end_time is None or time.time() < end_time
    while should_continue_iter and should_continue_time:
        count += 1
        last_poll_time = time.time()
        yield count
        sleep_till(last_poll_time + interval_secs)
        should_continue_iter = max_iter is None or count < max_iter
        should_continue_time = end_time is None or time.time() < end_time


def is_anyscale_workspace() -> bool:
    return "ANYSCALE_EXPERIMENTAL_WORKSPACE_ID" in os.environ


def credentials_check_sanity(credentials_str: str) -> bool:
    """
    The main goal of this function is to perform a minimal sanity check
    to make sure that the CLI token string (entered by user or read from file)
    is not totally broken.
    refer to https://www.notion.so/anyscale-hq/Authentication-Infrastructure
    """
    # Old token style
    if credentials_str.startswith("sss_"):
        return True
    # Future token style
    if credentials_str.startswith("a") and credentials_str.count("_") > 0:
        return True
    return False


def get_current_cluster_id() -> Optional[str]:
    """If we are running on an Anyscale Cluster, return the id from the environment."""
    return os.getenv("ANYSCALE_SESSION_ID")


def str_data_size(s: str) -> int:
    """Returns the size of the string when encoded to raw bytes."""
    return len(s.encode("utf-8"))



# AUTOGENERATED - modify shared_anyscale_util in root directory to make changes
import os


TEST_MODE = os.environ.get("TEST_MODE") == "1"

ANYSCALE_ENDPOINTS = {
    "development": "https://console.anyscale-dev.dev",
    "staging": "https://console.anyscale-staging.com",
    "production": "https://console.anyscale.com",
    "test": "",
    "integration_test": "localhost",
}

ANYSCALE_EXTERNAL_API_ENDPOINTS = {
    "development": "https://anyscale-dev.dev/ext",
    "staging": "https://api.anyscale-staging.com",
    "production": "https://api.anyscale.com",
    "test": "",
    "integration_test": "localhost",
}


if (
    "ANYSCALE_HOST" in os.environ
    and os.environ.get("ANYSCALE_HOST") not in ANYSCALE_ENDPOINTS.values()
):
    anyscale_env_default = "test"
else:
    anyscale_env_default = "production"

ANYSCALE_ENV = os.environ.get("DEPLOY_ENVIRONMENT", anyscale_env_default)
ANYSCALE_HOST = os.environ.get("ANYSCALE_HOST", ANYSCALE_ENDPOINTS[ANYSCALE_ENV])
ANYSCALE_API_HOST = ANYSCALE_EXTERNAL_API_ENDPOINTS[ANYSCALE_ENV]

RAY_STATIC_GCS_PORT = 9031

# The name of the default Anyscale shared k8s cloud.
ANYSCALE_K8S_CLOUD_NAME = "anyscale_k8s_cloud"
ANYSCALE_K8S_CLOUD_ID = "cld_41THJUKS3PBYvwIynRggIb"
ANYSCALE_K8S_GCP_CLOUD_NAME = "anyscale_k8s_gcp_cloud"
ANYSCALE_K8S_GCP_CLOUD_ID = "cld_k8WcxPgjUtSE8RVmfZpTLuKM"
AIOA_AWS_K8S_CLOUD_ID = "cld_HSrCZdMCYDe1NmMCJhYRgQ4p"

ANYSCALE_AWS_SECURITY_GROUP_NAME = "anyscale-security-group"

# env variable used to pass the AnyscaleBackgroundContext for background mode
ANYSCALE_BACKGROUND_JOB_CONTEXT = "ANYSCALE_BACKGROUND_JOB_CONTEXT"

# outer job name for background mode
OUTER_JOB_NAME = "_as_bg_job"

# The name of the BackgroundJobRunner actor
BACKGROUND_ACTOR_NAME = "_as_bg_actor"



import json
import time
from typing import Any
from unittest.mock import patch

import pytest

from anyscale.util import (
    _check_python_version,
    _ray_version_major_minor,
    _update_external_ids_for_policy,
    extract_versions_from_image_name,
    poll,
    sleep_till,
    str_data_size,
    updating_printer,
)


def test_updating_printer() -> None:
    out = ""

    def mock_print(
        string: str, *args: Any, end: str = "\n", flush: bool = False, **kwargs: Any
    ) -> None:
        nonlocal out
        out += string
        out += end

    with patch("anyscale.util.print", new=mock_print), patch(
        "shutil.get_terminal_size"
    ) as get_terminal_size_mock:
        get_terminal_size_mock.return_value = (10, 24)
        with updating_printer() as print_status:
            print_status("Step 1")
            print_status("Step 2")
            print_status("Step 3")

    assert out == (
        "\r          \r"
        "Step 1"
        "\r          \r"
        "Step 2"
        "\r          \r"
        "Step 3"
        "\r          \r"
    )


def test_updating_printer_multiline() -> None:
    out = ""

    def mock_print(
        string: str, *args: Any, end: str = "\n", flush: bool = False, **kwargs: Any
    ) -> None:
        nonlocal out
        out += string
        out += end

    with patch("anyscale.util.print", new=mock_print), patch(
        "shutil.get_terminal_size"
    ) as get_terminal_size_mock:
        get_terminal_size_mock.return_value = (10, 24)
        with updating_printer() as print_status:
            print_status("Step 1\nExtra stuff")
            print_status("ExtraLongLine12345")
            print_status("ExtraLongLine12345\nExtra stuff")
            print_status("Step 3")

    assert out == (
        "\r          \r"
        "Step 1..."
        "\r          \r"
        "ExtraLo..."
        "\r          \r"
        "ExtraLo..."
        "\r          \r"
        "Step 3"
        "\r          \r"
    )


STATEMENT_TEMPLATE = {
    "Action": "sts:AssumeRole",
    "Effect": "Allow",
    "Principal": {"AWS": "arn:aws:iam::ACCT_ID:root"},
}


@pytest.mark.parametrize(
    "statement_policy,expected_conditions",
    [
        pytest.param(
            [STATEMENT_TEMPLATE],
            [{"StringEquals": {"sts:ExternalId": ["new_id"]}}],
            id="OneStatement,NoPrior",
        ),
        pytest.param(
            [STATEMENT_TEMPLATE, STATEMENT_TEMPLATE],
            [{"StringEquals": {"sts:ExternalId": ["new_id"]}}] * 2,
            id="TwoStatements,NoPrior",
        ),
        pytest.param(
            [
                {
                    "Condition": {"StringEquals": {"sts:ExternalId": "old_id"}},
                    **STATEMENT_TEMPLATE,  # type: ignore
                }
            ],
            [{"StringEquals": {"sts:ExternalId": ["old_id", "new_id"]}}],
            id="OneStatement,OnePriorExternal",
        ),
        pytest.param(
            [
                {
                    "Condition": {"StringEquals": {"sts:ExternalId": "old_id"}},
                    **STATEMENT_TEMPLATE,  # type: ignore
                },
                STATEMENT_TEMPLATE,
            ],
            [
                {"StringEquals": {"sts:ExternalId": ["old_id", "new_id"]}},
                {"StringEquals": {"sts:ExternalId": ["new_id"]}},
            ],
            id="TwoStatements,OnePriorExternal",
        ),
        pytest.param(
            [
                {
                    "Condition": {"StringNotEquals": {"sts:ExternalId": "old_id"}},
                    **STATEMENT_TEMPLATE,  # type: ignore
                },
                STATEMENT_TEMPLATE,
            ],
            [
                {
                    "StringEquals": {"sts:ExternalId": ["new_id"]},
                    "StringNotEquals": {"sts:ExternalId": "old_id"},
                },
            ],
            id="OneStatemnt,OtherCondition",
        ),
    ],
)
def test_update_external_ids_for_policy(statement_policy, expected_conditions):
    policy_document = {
        "Statement": statement_policy,
        "Version": "2012-10-17",
    }
    new_policy = _update_external_ids_for_policy(policy_document, "new_id")

    for new, expected in zip(new_policy["Statement"], expected_conditions):
        assert new["Condition"] == expected


@pytest.mark.parametrize(
    "image_name, expected, exception_substr",
    [
        ("anyscale/ray-ml:1.11.1-py38-gpu", ("py38", "1.11.1"), None),
        ("anyscale/ray:1.12-py37-cpu", ("py37", "1.12"), None),
        ("anyscale/ray:1.12-py37", ("py37", "1.12"), None),
        ("anyscale/ray:1.12py37", None, "got 1.12py37"),
    ],
)
def test_extract_versions_from_image_name(image_name, expected, exception_substr):
    if exception_substr is not None:
        with pytest.raises(ValueError) as exc_info:
            extract_versions_from_image_name(image_name)
        assert exception_substr in str(exc_info.value)
    else:
        python_version, ray_version = extract_versions_from_image_name(image_name)
        assert (python_version, ray_version) == expected


@pytest.mark.parametrize(
    "ray_version, expected, exception_substr",
    [
        ("1.12", (1, 12), None),
        ("0.0", (0, 0), None),
        ("0:0", (0, 0), "unexpected"),
        ("112", (0, 0), "unexpected"),
        ("", (0, 0), "unexpected"),
        ("1.x", (0, 0), "unexpected"),
        ("0x10.12", (0, 0), "unexpected"),
    ],
)
def test_ray_version_major_minor(ray_version, expected, exception_substr):
    if exception_substr is not None:
        with pytest.raises(Exception) as exc_info:
            _ray_version_major_minor(ray_version)
        assert exception_substr in str(exc_info.value)
    else:
        got = _ray_version_major_minor(ray_version)
        assert got == expected


@pytest.mark.parametrize(
    "python_version, exception_substr",
    [
        ("py36", None),
        ("py37", None),
        ("py38", None),
        ("py39", None),
        ("py10", "got py10"),
        ("py3.6", "got py3.6"),
        ("py3", "got py3."),
        ("py35", "got py35"),
    ],
)
def test_python_version_major_minor(python_version, exception_substr):
    if exception_substr is not None:
        with pytest.raises(Exception) as exc_info:
            _check_python_version(python_version)
        assert exception_substr in str(exc_info.value)
    else:
        _check_python_version(python_version)


def test_poll():
    """Test the poll function.
    """

    end_time = time.time() + 0.5
    sleep_till(end_time)
    assert time.time() == pytest.approx(end_time)

    # This should poll forever
    count = 0
    start_time = time.time()
    for i in poll(0.01):
        count += 1
        assert count == i
        if count > 100:
            break
    assert count == 101
    assert time.time() == pytest.approx(start_time + 1.01)

    # Assert we stop iterating at max iter
    expected_i = 1
    for i in poll(0.01, max_iter=5):
        assert i == expected_i
        expected_i += 1
        assert i <= 5


def test_str_data_size():
    assert str_data_size("abcd") == 4

    # Serialized form: '{"hi": "ih"}'
    assert str_data_size(json.dumps({"hi": "ih"})) == 12



from typing import IO, Optional

import click

from anyscale.controllers.cluster_compute_controller import ClusterComputeController
from anyscale.util import validate_non_negative_arg


@click.group(
    "cluster-compute",
    short_help="Manage cluster compute configurations on Anyscale.",
    help="Manages cluster compute configurations to define cloud resource types and limitations.",
)
def cluster_compute_cli() -> None:
    pass


@cluster_compute_cli.command(
    name="create",
    help="Builds a new cluster compute template. For an example of a cluster compute file, see the example_compute_config.json created by running anyscale init in a project directory.",
)
@click.argument("cluster-compute-file", type=click.File("rb"), required=True)
@click.option(
    "--name",
    "-n",
    help="Name for the created cluster compute.",
    required=False,
    type=str,
)
def create_cluster_compute(
    cluster_compute_file: IO[bytes], name: Optional[str]
) -> None:
    ClusterComputeController().create(cluster_compute_file, name)


@cluster_compute_cli.command(
    name="delete", help="Delete the specified cluster compute template.",
)
@click.argument("cluster-compute-name", type=str, required=False)
@click.option(
    "--name",
    "-n",
    help="Name of the cluster compute template to delete.",
    required=False,
    type=str,
)
@click.option(
    "--cluster-compute-id",
    "--id",
    help="Id of the cluster compute template to delete. Must be provided if a cluster compute name is not given.",
    required=False,
    type=str,
)
def delete_cluster_compute(
    cluster_compute_name: Optional[str],
    name: Optional[str],
    cluster_compute_id: Optional[str],
) -> None:
    if cluster_compute_name is not None and name is not None:
        raise click.ClickException(
            "Please only provide one of [CLUSTER_COMPUTE_NAME] or --name."
        )
    ClusterComputeController().delete(cluster_compute_name or name, cluster_compute_id)


@cluster_compute_cli.command(
    name="list",
    help=(
        "List information about cluster computes on Anyscale. By default only list "
        "cluster computes you have created."
    ),
)
@click.option(
    "--name",
    "-n",
    required=False,
    default=None,
    help="List information about the cluster compute with this name.",
)
@click.option(
    "--cluster-compute-id",
    "--id",
    required=False,
    default=None,
    help=("List information about the cluster compute with this id."),
)
@click.option(
    "--include-shared",
    is_flag=True,
    default=False,
    help="Include all cluster cluster computes you have access to.",
)
@click.option(
    "--max-items",
    required=False,
    default=20,
    type=int,
    help="Max items to show in list.",
    callback=validate_non_negative_arg,
)
def list(
    name: Optional[str],
    cluster_compute_id: Optional[str],
    include_shared: bool,
    max_items: int,
) -> None:
    cluster_compute_controller = ClusterComputeController()
    cluster_compute_controller.list(
        cluster_compute_name=name,
        cluster_compute_id=cluster_compute_id,
        include_shared=include_shared,
        max_items=max_items,
    )


@cluster_compute_cli.command(
    name="get", help=("Get details about cluster compute configuration."),
)
@click.argument("cluster-compute-name", required=False)
@click.option(
    "--cluster-compute-id",
    "--id",
    required=False,
    default=None,
    help=("Get details about cluster compute configuration by this id."),
)
def get(cluster_compute_name: Optional[str], cluster_compute_id: Optional[str]) -> None:
    cluster_compute_controller = ClusterComputeController()
    cluster_compute_controller.get(
        cluster_compute_name=cluster_compute_name,
        cluster_compute_id=cluster_compute_id,
    )



# Used for data-gplane role.
AMAZON_S3_FULL_ACCESS_POLICY_ARN = "arn:aws:iam::aws:policy/AmazonS3FullAccess"

AMAZON_ECR_READONLY_ACCESS_POLICY_ARN = (
    "arn:aws:iam::aws:policy/AmazonEC2ContainerRegistryReadOnly"
)

# Used for control-plane role.
ANYSCALE_IAM_PERMISSIONS_EC2_STEADY_STATE = {
    "Version": "2012-10-17",
    "Statement": [
        {
            "Sid": "IAM",
            "Effect": "Allow",
            "Action": ["iam:PassRole", "iam:GetInstanceProfile"],
            "Resource": "*",
        },
        {
            "Sid": "RetrieveGenericAWSResources",
            "Effect": "Allow",
            "Action": [
                # Populates metadata about what is available
                # in the account.
                "ec2:DescribeAvailabilityZones",
                "ec2:DescribeInstanceTypes",
                "ec2:DescribeRegions",
            ],
            "Resource": "*",
        },
        {
            "Sid": "DescribeRunningResources",
            "Effect": "Allow",
            "Action": [
                # Determines cluster/configuration status.
                "ec2:DescribeInstances",
                "ec2:DescribeSubnets",
                "ec2:DescribeRouteTables",
                "ec2:DescribeSecurityGroups",
            ],
            "Resource": "*",
        },
        {
            "Sid": "InstanceManagementCore",
            "Effect": "Allow",
            "Action": [
                # Minimal Permissions to Run Instances on Anyscale.
                "ec2:RunInstances",
                "ec2:StartInstances",
                "ec2:StopInstances",
                "ec2:TerminateInstances",
            ],
            "Resource": "*",
        },
        {
            "Sid": "InstanceTagMangement",
            "Effect": "Allow",
            "Action": ["ec2:CreateTags", "ec2:DeleteTags"],
            "Resource": "*",
        },
        {
            "Sid": "InstanceManagementSpot",
            "Effect": "Allow",
            "Action": [
                # Extended Permissions to Run Instances on Anyscale.
                "ec2:CancelSpotInstanceRequests",
                "ec2:ModifyImageAttribute",
                "ec2:ModifyInstanceAttribute",
                "ec2:RequestSpotInstances",
            ],
            "Resource": "*",
        },
        {
            "Sid": "ResourceManagementExtended",
            "Effect": "Allow",
            "Action": [
                # Volume management
                "ec2:AttachVolume",
                "ec2:CreateVolume",
                "ec2:DeleteVolume",
                "ec2:DescribeVolumes",
                # IAMInstanceProfiles
                "ec2:AssociateIamInstanceProfile",
                "ec2:DisassociateIamInstanceProfile",
                "ec2:ReplaceIamInstanceProfileAssociation",
                # Placement groups
                "ec2:CreatePlacementGroup",
                "ec2:DeletePlacementGroup",
                # Address Management
                "ec2:AllocateAddress",
                "ec2:ReleaseAddress",
                # Additional DescribeResources
                "ec2:DescribeIamInstanceProfileAssociations",
                "ec2:DescribeInstanceStatus",
                "ec2:DescribePlacementGroups",
                "ec2:DescribePrefixLists",
                "ec2:DescribeReservedInstancesOfferings",
                "ec2:DescribeSpotInstanceRequests",
                "ec2:DescribeSpotPriceHistory",
            ],
            "Resource": "*",
        },
    ],
}

ANYSCALE_IAM_PERMISSIONS_EC2_INITIAL_RUN = {
    "Version": "2012-10-17",
    "Statement": [
        {
            "Sid": "SetupEC2",
            "Effect": "Allow",
            "Action": [
                # Anyscale runs this on the first time a cloud is configured.
                "ec2:CreateSecurityGroup",
                "ec2:AuthorizeSecurityGroupIngress",
                "ec2:AuthorizeSecurityGroupEgress",
                # For configuring VPCs
                "ec2:DescribeVpcs",
                "ec2:CreateVpc",
                "ec2:ModifyVpcAttribute",
                "ec2:CreateVpcEndpoint",
                # Add subnets
                "ec2:CreateSubnet",
                "ec2:ModifySubnetAttribute",
                # Add InternetGateway
                "ec2:CreateInternetGateway",
                "ec2:AttachInternetGateway",
                "ec2:DescribeInternetGateways",
                # Connect InternetGateway to Internet
                "ec2:CreateRouteTable",
                "ec2:AssociateRouteTable",
                "ec2:CreateRoute",
                "ec2:ReplaceRoute",
                # NAT Gateway Setup
                "ec2:CreateNatGateway",
                "ec2:DescribeNatGateways",
            ],
            "Resource": "*",
        },
        {
            "Sid": "CleanupEC2",
            "Effect": "Allow",
            "Action": [
                # Anyscale runs this on the first time a cloud is configured.
                "ec2:DeleteSecurityGroup",
                "ec2:RevokeSecurityGroupIngress",
                "ec2:RevokeSecurityGroupEgress",
                # Remove VPC
                "ec2:DeleteVpc",
                "ec2:DeleteVpcEndpoints",
                # Remove subnets
                "ec2:DeleteSubnet",
                # Remove InternetGateway
                "ec2:DeleteInternetGateway",
                "ec2:DetachInternetGateway",
                # Disconnect InternetGateway
                "ec2:DeleteRouteTable",
                "ec2:DisassociateRouteTable",
                "ec2:DeleteRoute",
                # Remove NATGateway
                "ec2:DeleteNatGateway",
            ],
            "Resource": "*",
        },
    ],
}

# Used for experimental readonly access to SSM
ANYSCALE_SSM_READONLY_ACCESS_POLICY_DOCUMENT = {
    "Version": "2012-10-17",
    "Statement": {
        "Sid": "1",
        "Effect": "Allow",
        "Action": [
            "secretsmanager:GetSecretValue",
            "secretsmanager:DescribeSecret",
            "secretsmanager:ListSecrets",
        ],
        "Resource": "*",
    },
}

# Used for experimental read and write access to SSM
ANYSCALE_SSM_READ_WRITE_ACCESS_POLICY_DOCUMENT = {
    "Version": "2012-10-17",
    "Statement": {
        "Sid": "1",
        "Effect": "Allow",
        "Action": [
            "secretsmanager:CreateSecret",
            "secretsmanager:PutSecretValue",
            "secretsmanager:GetSecretValue",
            "secretsmanager:DescribeSecret",
            "secretsmanager:ListSecrets",
        ],
        "Resource": "*",
    },
}



import os
from pathlib import Path
import random
import shutil
import tempfile
import uuid
from zipfile import ZipFile

import pytest
import smart_open
import yaml

from anyscale.sdk.anyscale_client.models import CreateProductionJob
from anyscale.sdk.anyscale_client.sdk import (
    _upload_and_rewrite_working_dir_in_create_production_job,
)
from anyscale.utils.ray_utils import (  # type: ignore
    _dir_travel,
    _get_excludes,
    zip_directory,
)
from anyscale.utils.runtime_env import (
    _get_remote_storage_object_name,
    upload_and_rewrite_working_dir,
)


CONDA_DICT = {"dependencies": ["pip", {"pip": ["pip-install-test==0.5"]}]}
PIP_LIST = ["requests==1.0.0", "pip-install-test"]
GS_BUCKET_NAME = "anyscale-bk-e2e-test-job"


@pytest.fixture
def test_directory():
    with tempfile.TemporaryDirectory() as tmp_dir:
        path = Path(tmp_dir)
        subdir = path / "subdir"
        subdir.mkdir(parents=True)
        requirements_file = subdir / "requirements.txt"
        with requirements_file.open(mode="w") as f:
            print("\n".join(PIP_LIST), file=f)

        good_conda_file = subdir / "good_conda_env.yaml"
        with good_conda_file.open(mode="w") as f:
            yaml.dump(CONDA_DICT, f)

        bad_conda_file = subdir / "bad_conda_env.yaml"
        with bad_conda_file.open(mode="w") as f:
            print("% this is not a YAML file %", file=f)

        yield subdir, requirements_file, good_conda_file, bad_conda_file


@pytest.mark.parametrize(
    "upload_path,expected_output",
    [
        ("gs://bucket", "file.zip"),
        ("gs://bucket/path", "path/file.zip"),
        ("gs://bucket/path/", "path/file.zip"),
    ],
)
def test_get_remote_storage_object_name(upload_path, expected_output):
    assert (
        _get_remote_storage_object_name(upload_path, upload_filename="file.zip")
        == expected_output
    )


class TestUploadLocalWorkingDir:
    def test_upload_and_rewrite_working_dir_with_fake_upload(self, test_directory):
        def fake_upload_file_to_remote_storage(
            source_file: str, upload_path: str, upload_filename: str
        ) -> str:
            shutil.copyfile(source_file, upload_filename)
            return os.path.join(upload_path, upload_filename)

        upload_path = "s3://fake/dir"
        runtime_env = {
            "pip": ["blah"],
            "working_dir": str(test_directory),
            "upload_path": upload_path,
        }
        new_runtime_env = upload_and_rewrite_working_dir(
            runtime_env,
            upload_file_to_remote_storage_fn=fake_upload_file_to_remote_storage,
        )

        assert new_runtime_env["working_dir"].startswith(upload_path)
        assert "upload_path" not in new_runtime_env
        assert Path(new_runtime_env["working_dir"]).suffix == ".zip"

    @pytest.mark.skipif(
        os.environ.get("CI") != "true",
        reason="Skip test if not running in CI because S3/GS credentials are not available",
    )
    @pytest.mark.parametrize(
        "upload_path",
        [
            "s3://bk-premerge-first-jawfish-artifacts/e2e_tests/job",
            f"gs://{GS_BUCKET_NAME}",
        ],
    )
    @pytest.mark.parametrize("mode", ["sdk", "cli"])
    def test_upload_and_rewrite_working_dir(
        self, test_directory, upload_path: str, mode
    ):
        """Test that the working_dir correctly gets uploaded to upload_path in the runtime_env."""
        subdir, _, _, _ = test_directory
        runtime_env = {
            "pip": ["blah"],
            "working_dir": str(subdir),
            "upload_path": upload_path,
        }
        if mode == "sdk":
            job_config = {
                "compute_config_id": "cpt_U8RCfD7Wr1vCD4iqGi4cBbj1",
                "build_id": "bld_1277XIinoJmiM8Z3gNdcHN",
                "runtime_env": runtime_env,
                "entrypoint": "python my_job_script.py --option1=value1",
                "max_retries": 3,
            }

            create_production_job = CreateProductionJob(
                name="my-production-job",
                description="A production job running on Anyscale.",
                project_id="prj_7S7Os7XBvO6vdiVC1J0lgj",
                config=job_config,
            )

            _upload_and_rewrite_working_dir_in_create_production_job(
                create_production_job
            )
            new_runtime_env = create_production_job.config.get("runtime_env")
        elif mode == "cli":
            # Upload the working_dir and rewrite the working_dir field to point to the new remote URI
            new_runtime_env = upload_and_rewrite_working_dir(runtime_env)
        else:
            raise ValueError(f"Invalid mode: {mode}")

        # Check all non-working_dir fields in the runtime_env are unchanged
        runtime_env_copy = runtime_env.copy()
        new_runtime_env_copy = new_runtime_env.copy()
        runtime_env_copy.pop("working_dir")
        runtime_env_copy.pop("upload_path")
        new_runtime_env_copy.pop("working_dir")
        assert runtime_env_copy == new_runtime_env_copy

        # Check the format of the remote URI
        remote_uri = new_runtime_env["working_dir"]
        assert remote_uri.startswith(upload_path)
        assert Path(remote_uri).suffix == ".zip"

        # Download the zip file from the remote bucket
        local_zip_file = "test_job_controller_working_dir.zip"
        with smart_open.open(remote_uri, "rb") as package_zip:
            with open(local_zip_file, "wb") as fin:
                fin.write(package_zip.read())

        # Check the downloaded zip is the same directory that was uploaded
        unzip_target_dir = os.path.join("unzip_target_dir")
        with ZipFile(local_zip_file, "r") as zip_ref:
            zip_ref.extractall(unzip_target_dir)
        assert set(os.listdir(os.path.join(unzip_target_dir, "subdir"))) == {
            "requirements.txt",
            "good_conda_env.yaml",
            "bad_conda_env.yaml",
        }

    def test_remote_uri_working_dir_noop(self):
        runtime_env = {"working_dir": "s3://already-a-uri"}
        assert runtime_env == upload_and_rewrite_working_dir(runtime_env)

    @pytest.mark.parametrize("include_parent_dir", [True, False])
    def test_zip_directory(self, test_directory, include_parent_dir):
        subdir, _, _, _ = test_directory
        with tempfile.TemporaryDirectory() as tmpdir:
            zip_file = os.path.join(tmpdir, "test_zip.zip")
            # NOTE(architkulkarni): Must use "requirements.txt" in excludes instead of requirements_file
            # until https://github.com/ray-project/ray/issues/23473 is fixed.
            zip_directory(
                subdir,
                excludes=["requirements.txt"],
                output_path=zip_file,
                include_parent_dir=include_parent_dir,
            )
            unzip_target_dir = os.path.join(tmpdir, "unzip_target_dir")
            with ZipFile(zip_file, "r") as zip_ref:
                zip_ref.extractall(unzip_target_dir)
            if include_parent_dir:
                assert os.listdir(unzip_target_dir) == ["subdir"]
                unzip_target_dir = os.path.join(unzip_target_dir, "subdir")
            assert set(os.listdir(unzip_target_dir)) == {
                "good_conda_env.yaml",
                "bad_conda_env.yaml",
            }

    def test_travel(self, tmp_path):
        """Copied from OSS Ray at python/ray/tests/test_runtime_env_packaging.py"""
        dir_paths = set()
        file_paths = set()
        item_num = 0
        excludes = []
        root = tmp_path / "test"

        def construct(path, excluded=False, depth=0):
            nonlocal item_num
            path.mkdir(parents=True)
            if not excluded:
                dir_paths.add(str(path))
            if depth > 8:
                return
            if item_num > 500:
                return
            dir_num = random.randint(0, 10)
            file_num = random.randint(0, 10)
            for _ in range(dir_num):
                uid = str(uuid.uuid4()).split("-")[0]
                dir_path = path / uid
                exclud_sub = random.randint(0, 5) == 0
                if not excluded and exclud_sub:
                    excludes.append(str(dir_path.relative_to(root)))
                if not excluded:
                    construct(dir_path, exclud_sub or excluded, depth + 1)
                item_num += 1
            if item_num > 1000:
                return

            for _ in range(file_num):
                uid = str(uuid.uuid4()).split("-")[0]
                v = random.randint(0, 1000)
                with (path / uid).open("w") as f:
                    f.write(str(v))
                if not excluded:
                    if random.randint(0, 5) == 0:
                        excludes.append(str((path / uid).relative_to(root)))
                    else:
                        file_paths.add((str(path / uid), str(v)))
                item_num += 1

        construct(root)
        exclude_spec = _get_excludes(root, excludes)
        visited_dir_paths = set()
        visited_file_paths = set()

        def handler(path):
            if path.is_dir():
                visited_dir_paths.add(str(path))
            else:
                with open(path) as f:
                    visited_file_paths.add((str(path), f.read()))

        _dir_travel(root, [exclude_spec], handler)
        assert file_paths == visited_file_paths
        assert dir_paths == visited_dir_paths


@pytest.mark.skipif(
    os.environ.get("CI") != "true",
    reason="Skip test if not running in CI because credentials are not available.",
)
def test_gs_upload_download():
    """Test that the CI machine has access to the Google Storage bucket used for testing."""
    random_int = random.randint(0, 100000)
    filename = f"test_gs_upload_download_{random_int}.txt"

    from google.cloud import storage

    storage_client = storage.Client()
    bucket_obj = storage_client.bucket(GS_BUCKET_NAME)
    blob = bucket_obj.blob(filename)
    try:
        random_str = str(random_int)
        blob.upload_from_string(random_str)
        assert blob.exists()

        from smart_open import open

        uri = f"gs://{GS_BUCKET_NAME}/{filename}"

        with open(uri, "r") as f:
            assert random_str == f.read()
    finally:
        # Delete the file from the Google Storage bucket
        blob.delete()



import json
import logging
import os
import traceback
from typing import Any, Dict, Optional, Tuple

from anyscale.authenticate import get_auth_api_client
from anyscale.util import str_data_size


logger = logging.getLogger("Job Outputs")

ANYSCALE_HA_JOB_ID = "ANYSCALE_HA_JOB_ID"
ANYSCALE_RAY_JOB_SUBMISSION_ID = "ANYSCALE_RAY_JOB_SUBMISSION_ID"


def output(result: Dict[str, Any], *, _api_client=None, _fail_on_error=False):
    """Submit the output for an Anyscale Production Job

    Outputs can only be recorded from within an Anyscale Production Job.

    This function should be called from the driver of your ray script.

    Example usage for this function can be found in the Anyscale documentation.

    Args:
        result (Dict[str, Any]): A JSON serializable dictionary representing your output data. Size limit is 1MiB.

    Raises:
        e: if the internal flag _fail_on_error is set, this function will raise any exceptions from the API
    """
    try:
        if not isinstance(result, dict):
            logger.warn(
                "The output of your function must be a dictionary. Your structured output will not be recorded."
            )
            return

        try:
            serialized_result = json.dumps(result, indent=2)
        except Exception:
            traceback.print_exc()
            logger.warn(
                "Failed to serialize the job output to JSON. Output will not be recorded."
            )
            return

        if str_data_size(serialized_result) > 1024 ** 2:
            logger.warn(
                "Job output exceeds maximum size of 1MiB. Output will not be recorded."
            )
            return

        # Check if we're in a production job last so we still go through the above checks
        # to aid users with debugging.
        ids = _get_ha_job_id_from_environment()
        if ids is None:
            logger.warn(
                "Job outputs can only be recorded from Production Jobs. Output will be ignored."
            )
            return

        ha_job_id, ray_job_submission_id = ids
        _submit_raw_output(
            ha_job_id, ray_job_submission_id, result, api_client=_api_client
        )
    except Exception as e:
        # Only crash the program if we have explicitly set the flag
        if _fail_on_error:
            raise e from None
        else:
            logger.warn(
                f"Your structured output was not recorded because of an error: {e}"
            )


def _submit_raw_output(
    ha_job_id: str,
    ray_job_submission_id: str,
    output: Dict[str, Any],
    *,
    api_client=None,
):
    if api_client is None:
        api_client = get_auth_api_client().api_client
    return api_client.create_structured_output_api_v2_structured_outputs_post(
        {
            "production_job_id": ha_job_id,
            "output": output,
            "ray_job_submission_id": ray_job_submission_id,
        }
    )


def _get_ha_job_id_from_environment() -> Optional[Tuple[str, str]]:
    ha_job_id = os.getenv(ANYSCALE_HA_JOB_ID)
    ray_job_submission_id = os.getenv(ANYSCALE_RAY_JOB_SUBMISSION_ID)

    if not ha_job_id:
        logger.debug(f"{ANYSCALE_HA_JOB_ID} env var not set.")

    if not ray_job_submission_id:
        logger.debug(f"{ANYSCALE_RAY_JOB_SUBMISSION_ID} env var not set.")

    if ha_job_id and ray_job_submission_id:
        return ha_job_id, ray_job_submission_id

    return None



from typing import Any, Dict, Iterator, List, Optional, Tuple
from unittest.mock import ANY, Mock, patch

from click import ClickException
import pytest

from anyscale.aws_iam_policies import (
    AMAZON_ECR_READONLY_ACCESS_POLICY_ARN,
    AMAZON_S3_FULL_ACCESS_POLICY_ARN,
)
from anyscale.client.openapi_client.models import (
    AnyscaleAWSAccount,
    AnyscaleawsaccountResponse,
    Cloud,
    CloudConfig,
    CloudResponse,
)
from anyscale.controllers.cloud_controller import CloudController


@pytest.fixture()
def mock_api_client(cloud_test_data: Cloud) -> Mock:
    mock_api_client = Mock()
    mock_api_client.get_cloud_api_v2_clouds_cloud_id_get = Mock(
        return_value=CloudResponse(result=cloud_test_data)
    )
    mock_api_client.find_cloud_by_name_api_v2_clouds_find_by_name_post = Mock(
        return_value=CloudResponse(result=cloud_test_data)
    )
    mock_api_client.delete_cloud_api_v2_clouds_cloud_id_delete = Mock(return_value={})
    mock_api_client.update_cloud_config_api_v2_clouds_cloud_id_config_put = Mock(
        return_value={}
    )

    return mock_api_client


@pytest.fixture(autouse=True)
def mock_auth_api_client(
    mock_api_client: Mock, base_mock_anyscale_api_client: Mock
) -> Iterator[None]:
    mock_auth_api_client = Mock(
        api_client=mock_api_client, anyscale_api_client=base_mock_anyscale_api_client,
    )
    with patch.multiple(
        "anyscale.controllers.base_controller",
        get_auth_api_client=Mock(return_value=mock_auth_api_client),
    ):
        yield


@pytest.fixture(autouse=True)
def mock_get_available_regions() -> Iterator[None]:
    with patch(
        "anyscale.controllers.cloud_controller.get_available_regions",
        return_value=["us-west-2"],
        autospec=True,
    ):
        yield


def mock_role(document: Optional[Dict[str, Any]] = None) -> Mock:
    if document is None:
        document = {}
    mock_role = Mock()

    mock_role.arn = "ARN"
    mock_role.attach_policy = Mock()
    mock_role.assume_role_policy_document = document
    mock_role.instance_profiles.all = Mock(return_value=[])

    return mock_role


def mock_role_with_external_id() -> Mock:
    return mock_role(
        {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "1",
                    "Effect": "Allow",
                    "Principal": {"AWS": ["ARN"]},
                    "Action": "sts:AssumeRole",
                    "Condition": {"StringEquals": {"sts:ExternalId": "extid"}},
                }
            ],
        }
    )


def test_setup_cloud_aws() -> None:
    with patch.object(
        CloudController, "setup_aws", return_value=None
    ) as mock_setup_aws:
        cloud_controller = CloudController()
        cloud_controller.setup_cloud(
            provider="aws", region=None, name="test-aws", yes=False,
        )

        mock_setup_aws.assert_called_once_with(
            region="us-west-2", name="test-aws", yes=False
        )


def test_setup_cloud_gcp() -> None:
    mock_launch_gcp_cloud_setup = Mock(return_value=None)
    with patch.multiple(
        "anyscale.controllers.cloud_controller",
        launch_gcp_cloud_setup=mock_launch_gcp_cloud_setup,
    ):
        cloud_controller = CloudController()
        cloud_controller.setup_cloud(
            provider="gcp", region=None, name="test-gcp", yes=True,
        )

        mock_launch_gcp_cloud_setup.assert_called_once_with(
            region="us-west1",
            name="test-gcp",
            folder_id=None,
            vpc_peering_ip_range=None,
            vpc_peering_target_project_id=None,
            vpc_peering_target_vpc_id=None,
        )

        mock_launch_gcp_cloud_setup.reset_mock()

        cloud_controller.setup_cloud(
            provider="gcp", region=None, name="test-gcp", yes=True, folder_id=1234,
        )

        mock_launch_gcp_cloud_setup.assert_called_once_with(
            region="us-west1",
            name="test-gcp",
            folder_id=1234,
            vpc_peering_ip_range=None,
            vpc_peering_target_project_id=None,
            vpc_peering_target_vpc_id=None,
        )


@pytest.mark.parametrize(
    "vpc_peering_options",
    [
        ("10.0.0.0/12", "project_id", "vpc_id"),
        ("10.0.0.0/12", None, "vpc_id"),
        ("10.0.0.0/12", "project_id", None),
    ],
)
def test_setup_cloud_gcp_vpc_peering(
    vpc_peering_options: Tuple[str, Optional[str], Optional[str]]
) -> None:
    (
        vpc_peering_ip_range,
        vpc_peering_target_project_id,
        vpc_peering_target_vpc_id,
    ) = vpc_peering_options
    mock_launch_gcp_cloud_setup = Mock(return_value=None)
    with patch.multiple(
        "anyscale.controllers.cloud_controller",
        launch_gcp_cloud_setup=mock_launch_gcp_cloud_setup,
    ):
        cloud_controller = CloudController()
        if vpc_peering_target_project_id is None or vpc_peering_target_vpc_id is None:
            with pytest.raises(ClickException):
                cloud_controller.setup_cloud(
                    provider="gcp",
                    region=None,
                    name="test-gcp",
                    yes=True,
                    vpc_peering_ip_range=vpc_peering_ip_range,
                    vpc_peering_target_project_id=vpc_peering_target_project_id,
                    vpc_peering_target_vpc_id=vpc_peering_target_vpc_id,
                )
        else:
            cloud_controller.setup_cloud(
                provider="gcp",
                region=None,
                name="test-gcp",
                yes=True,
                vpc_peering_ip_range=vpc_peering_ip_range,
                vpc_peering_target_project_id=vpc_peering_target_project_id,
                vpc_peering_target_vpc_id=vpc_peering_target_vpc_id,
            )

        if vpc_peering_target_project_id is None or vpc_peering_target_vpc_id is None:
            return

        mock_launch_gcp_cloud_setup.assert_called_once_with(
            region="us-west1",
            name="test-gcp",
            folder_id=None,
            vpc_peering_ip_range=vpc_peering_ip_range,
            vpc_peering_target_project_id=vpc_peering_target_project_id,
            vpc_peering_target_vpc_id=vpc_peering_target_vpc_id,
        )

        mock_launch_gcp_cloud_setup.reset_mock()

        cloud_controller.setup_cloud(
            provider="gcp",
            region=None,
            name="test-gcp",
            yes=True,
            folder_id=1234,
            vpc_peering_ip_range=vpc_peering_ip_range,
            vpc_peering_target_project_id=vpc_peering_target_project_id,
            vpc_peering_target_vpc_id=vpc_peering_target_vpc_id,
        )

        mock_launch_gcp_cloud_setup.assert_called_once_with(
            region="us-west1",
            name="test-gcp",
            folder_id=1234,
            vpc_peering_ip_range=vpc_peering_ip_range,
            vpc_peering_target_project_id=vpc_peering_target_project_id,
            vpc_peering_target_vpc_id=vpc_peering_target_vpc_id,
        )


@pytest.mark.parametrize(
    "vpc_peering_options,error_output",
    [
        (
            ("not_an_ip_address", "project_id", "vpc_id"),
            "not_an_ip_address is not a valid IP address.",
        ),
        (
            ("10.0.0.0/20", "project_id", "vpc_id"),
            "10.0.0.0/20 is not a valid IP range. The minimum size is /16",
        ),
        (
            ("42.0.0.0/10", "project_id", "vpc_id"),
            "42.0.0.0/10 is not a allowed private IP address range for GCP",
        ),
    ],
)
def test_setup_cloud_gcp_vpc_peering_validation(
    vpc_peering_options: Tuple[str, str, str], error_output: str
) -> None:
    (
        vpc_peering_ip_range,
        vpc_peering_target_project_id,
        vpc_peering_target_vpc_id,
    ) = vpc_peering_options
    mock_launch_gcp_cloud_setup = Mock(return_value=None)
    with patch.multiple(
        "anyscale.controllers.cloud_controller",
        launch_gcp_cloud_setup=mock_launch_gcp_cloud_setup,
    ):
        cloud_controller = CloudController()

        with pytest.raises(ClickException) as e:
            cloud_controller.setup_cloud(
                provider="gcp",
                region=None,
                name="test-gcp",
                yes=True,
                vpc_peering_ip_range=vpc_peering_ip_range,
                vpc_peering_target_project_id=vpc_peering_target_project_id,
                vpc_peering_target_vpc_id=vpc_peering_target_vpc_id,
            )

        assert error_output in e.value.message


def test_setup_cloud_gcp_bad_region() -> None:
    mock_confirm = Mock(side_effect=ClickException("aborted"))
    with patch.multiple(
        "anyscale.controllers.cloud_controller", confirm=mock_confirm,
    ):
        cloud_controller = CloudController()
        # NOTE: GCP regions are [cont]-[local][number], not [cont]-[local]-[number]
        with pytest.raises(ClickException):
            cloud_controller.setup_cloud(
                provider="gcp", region="us-west-2", name="test-gcp",
            )

        mock_confirm.assert_called()


def test_setup_cloud_invalid_provider() -> None:
    cloud_controller = CloudController()
    with pytest.raises(ClickException):
        cloud_controller.setup_cloud(
            provider="azure",
            region="azure-west-1",
            name="invalid cloud provider",
            yes=False,
        )


def test_delete_cloud_by_name(cloud_test_data: Cloud) -> None:
    cloud_controller = CloudController()
    success = cloud_controller.delete_cloud(
        cloud_id=None, cloud_name=cloud_test_data.name, skip_confirmation=True
    )
    assert success

    cloud_controller.api_client.find_cloud_by_name_api_v2_clouds_find_by_name_post.assert_called_once_with(
        cloud_name_options={"name": cloud_test_data.name}
    )
    cloud_controller.api_client.delete_cloud_api_v2_clouds_cloud_id_delete(
        cloud_id=cloud_test_data.id
    )


def test_delete_cloud_by_id(cloud_test_data: Cloud) -> None:
    cloud_controller = CloudController()
    success = cloud_controller.delete_cloud(
        cloud_id=cloud_test_data.id, cloud_name=None, skip_confirmation=True
    )
    assert success

    cloud_controller.api_client.get_cloud_api_v2_clouds_cloud_id_get.assert_called_once_with(
        cloud_id=cloud_test_data.id
    )
    cloud_controller.api_client.delete_cloud_api_v2_clouds_cloud_id_delete(
        cloud_id=cloud_test_data.id
    )


def test_missing_name_and_id() -> None:
    cloud_controller = CloudController()

    with pytest.raises(ClickException):
        cloud_controller.delete_cloud(None, None, True)

    with pytest.raises(ClickException):
        cloud_controller.update_cloud_config(None, None, 0)

    with pytest.raises(ClickException):
        cloud_controller.get_cloud_config(None, None)


@pytest.mark.parametrize("role_name_exists", [True, False])
def test_setup_cross_region(cloud_test_data: Cloud, role_name_exists: bool) -> None:
    mock_get_aws_account = Mock(
        return_value=AnyscaleawsaccountResponse(
            result=AnyscaleAWSAccount(anyscale_aws_account="aws_account_type")
        )
    )
    mock_create_cloud = Mock(return_value=CloudResponse(result=cloud_test_data))

    mock_role = mock_role_with_external_id()
    mock_role_policy = Mock()
    mock_role.Policy.return_value = mock_role_policy
    if role_name_exists:
        roles = [mock_role, None, mock_role]
    else:
        roles = [None, mock_role]

    mock_iam = Mock()
    with patch.multiple(
        "anyscale.controllers.cloud_controller", _get_role=Mock(side_effect=roles),
    ), patch.multiple("boto3", resource=Mock(return_value=mock_iam),), patch.multiple(
        "secrets", token_hex=lambda x: "f" * x * 2
    ):
        cloud_controller = CloudController()
        cloud_controller.api_client.get_anyscale_aws_account_api_v2_clouds_anyscale_aws_account_get = (
            mock_get_aws_account
        )
        cloud_controller.api_client.create_cloud_api_v2_clouds_post = mock_create_cloud
        cloud_controller.setup_aws_cross_account_role("us-west-2", "name")

    cloud_controller.api_client.get_anyscale_aws_account_api_v2_clouds_anyscale_aws_account_get.assert_called_once()

    assert (
        mock_role_policy.put.call_count == 2
    ), "Expected 2 calls, only got {}:\n{}".format(
        mock_role_policy.put.call_count, mock_role_policy.put.mock_calls
    )
    mock_role.AssumeRolePolicy.assert_called()

    mock_iam.create_role.assert_called_once_with(
        AssumeRolePolicyDocument=ANY, RoleName="anyscale-iam-role-ffffffff"
    )
    mock_iam.meta.client.update_role.assert_called_once()


@pytest.mark.parametrize(
    "roles",
    [
        pytest.param([None, mock_role()], id="role_doesnt_exist"),
        pytest.param([mock_role()], id="role_already_exists"),
    ],
)
def test_setup_aws_ray_role(roles: List[Optional[Mock]]) -> None:
    assert roles[-1] is not None, "roles must end with a real role"

    mock_iam = Mock()
    mock_iam.create_role = Mock()

    with patch.multiple(
        "anyscale.controllers.cloud_controller", _get_role=Mock(side_effect=roles),
    ), patch.multiple(
        "boto3", resource=Mock(return_value=mock_iam),
    ):
        cloud_controller = CloudController()
        cloud_controller.setup_aws_ray_role("us-west-2", "ray-autoscaler-v1")

    if roles[0] is None:
        # Role didn't exist at the start and had to be "created"
        mock_iam.create_role.assert_called_once()

    # Assert we actually attached the base policies
    roles[-1].attach_policy.assert_any_call(PolicyArn=AMAZON_S3_FULL_ACCESS_POLICY_ARN)
    roles[-1].attach_policy.assert_any_call(
        PolicyArn=AMAZON_ECR_READONLY_ACCESS_POLICY_ARN
    )
    assert 2 == roles[-1].attach_policy.call_count
    mock_iam.create_instance_profile.assert_called_once_with(
        InstanceProfileName="ray-autoscaler-v1"
    )


def test_update_cloud_config_by_name(cloud_test_data: Cloud) -> None:
    cloud_controller = CloudController()
    cloud_controller.update_cloud_config(
        cloud_id=None, cloud_name=cloud_test_data.name, max_stopped_instances=100,
    )

    cloud_controller.api_client.find_cloud_by_name_api_v2_clouds_find_by_name_post.assert_called_once_with(
        cloud_name_options={"name": cloud_test_data.name}
    )
    cloud_controller.api_client.update_cloud_config_api_v2_clouds_cloud_id_config_put.assert_called_once_with(
        cloud_id=cloud_test_data.id,
        cloud_config=CloudConfig(max_stopped_instances=100),
    )


def test_update_cloud_config_by_id(cloud_test_data: Cloud) -> None:
    cloud_controller = CloudController()
    cloud_controller.update_cloud_config(
        cloud_id=cloud_test_data.id, cloud_name=None, max_stopped_instances=100,
    )

    cloud_controller.api_client.update_cloud_config_api_v2_clouds_cloud_id_config_put.assert_called_once_with(
        cloud_id=cloud_test_data.id,
        cloud_config=CloudConfig(max_stopped_instances=100),
    )


@pytest.mark.parametrize("cloud_id", [None, "cloud_id_1"])
@pytest.mark.parametrize("cloud_name", [None, "cloud_name_1"])
def test_set_default_cloud(cloud_id: Optional[str], cloud_name: Optional[str]) -> None:
    cloud_controller = CloudController()
    if not (cloud_id or cloud_name) or (cloud_id and cloud_name):
        # Error if neither or both of cloud_id and cloud_name provided
        with pytest.raises(ClickException):
            cloud_controller.set_default_cloud(
                cloud_id=cloud_id, cloud_name=cloud_name,
            )
    else:
        cloud_controller.set_default_cloud(
            cloud_id=cloud_id, cloud_name=cloud_name,
        )
        cloud_controller.api_client.update_default_cloud_api_v2_organizations_update_default_cloud_post.assert_called_once_with(
            cloud_id="cloud_id_1"
        )


@pytest.mark.parametrize("cloud_id", [None, "cloud_id_1"])
@pytest.mark.parametrize("cloud_name", [None, "cloud_name_1"])
def test_list_cloud(cloud_id: Optional[str], cloud_name: Optional[str]) -> None:
    cloud_controller = CloudController()
    cloud_controller.api_client.list_clouds_api_v2_clouds_get = Mock(
        return_value=Mock(results=[Mock()])
    )
    cloud_controller.list_clouds(cloud_name, cloud_id)

    if cloud_id is not None:
        cloud_controller.api_client.get_cloud_api_v2_clouds_cloud_id_get.assert_called_once_with(
            cloud_id
        )
    elif cloud_name is not None:
        cloud_controller.api_client.find_cloud_by_name_api_v2_clouds_find_by_name_post.assert_called_once_with(
            {"name": cloud_name}
        )
    else:
        cloud_controller.api_client.list_clouds_api_v2_clouds_get.assert_called_once_with()



import os
from typing import Optional
from unittest.mock import ANY, Mock, mock_open as mock_open_factory, patch

import pytest

from anyscale.client.openapi_client import Project, UserInfo
from anyscale.client.openapi_client.models.project_list_response import (
    ProjectListResponse,
)
from anyscale.controllers.project_controller import (
    COMPUTE_CONFIG_FILENAME,
    ProjectController,
)
from anyscale.project import ProjectDefinition


@pytest.fixture()
def mock_api_client(project_test_data: Project) -> Mock:
    mock_api_client = Mock()

    mock_api_client.find_project_by_project_name_api_v2_projects_find_by_name_get.return_value = ProjectListResponse(
        results=[project_test_data]
    )
    mock_api_client.list_sessions_api_v2_sessions_get.return_value.results = []
    mock_api_client.get_project_latest_cluster_config_api_v2_projects_project_id_latest_cluster_config_get.return_value.result.config = (
        ""
    )

    return mock_api_client


@pytest.fixture()
def mock_auth_api_client(mock_api_client: Mock, base_mock_anyscale_api_client: Mock):
    mock_auth_api_client = Mock(
        api_client=mock_api_client, anyscale_api_client=base_mock_anyscale_api_client,
    )
    with patch.multiple(
        "anyscale.controllers.base_controller",
        get_auth_api_client=Mock(return_value=mock_auth_api_client),
    ):
        yield


@pytest.mark.parametrize("owner", [None, "owner"])
def test_clone_project(
    project_test_data: Project, owner: Optional[str], mock_auth_api_client
) -> None:
    project_controller = ProjectController()
    project_controller._write_sample_compute_config = Mock()  # type: ignore

    os_makedirs_mock = Mock(return_value=None)
    mockwrite_project_file_to_disk = Mock(return_value=None)
    with patch.multiple("os", makedirs=os_makedirs_mock), patch(
        "anyscale.controllers.project_controller.write_project_file_to_disk"
    ) as mockwrite_project_file_to_disk:
        project_controller.clone(project_name=project_test_data.name, owner=owner)

    project_controller.api_client.find_project_by_project_name_api_v2_projects_find_by_name_get.assert_called_once_with(
        name=project_test_data.name, _request_timeout=ANY, owner=owner
    )
    os_makedirs_mock.assert_called_once_with(project_test_data.name)
    mockwrite_project_file_to_disk.assert_called_once_with(
        project_test_data.id, project_test_data.name
    )

    project_controller._write_sample_compute_config.assert_called_once_with(
        filepath=os.path.join(project_test_data.name, COMPUTE_CONFIG_FILENAME),
        project_id=project_test_data.id,
    )


def test_init_project(project_test_data: Project, mock_auth_api_client) -> None:
    project_controller = ProjectController()
    project_controller._write_sample_compute_config = Mock()  # type: ignore
    project_definition = ProjectDefinition(os.getcwd())
    project_definition.config["name"] = project_test_data.name
    mock_create_new_proj_def = Mock(
        return_value=(project_test_data.name, project_definition)
    )
    mock_register_project = Mock()
    os_path_exists_mock = Mock(return_value=False)

    with patch.multiple(
        "anyscale.controllers.project_controller",
        create_new_proj_def=mock_create_new_proj_def,
        register_or_attach_to_project=mock_register_project,
    ), patch.multiple("os.path", exists=os_path_exists_mock):
        project_controller.init(project_id=None, name=project_test_data.name)

    mock_create_new_proj_def.assert_called_once_with(
        project_test_data.name, api_client=project_controller.api_client,
    )
    mock_register_project.assert_called_once_with(
        project_definition, project_controller.api_client
    )


def test_init_project_with_id(project_test_data: Project, mock_auth_api_client) -> None:
    project_controller = ProjectController()
    project_controller._write_sample_compute_config = Mock()  # type: ignore
    project_definition = ProjectDefinition(os.getcwd())
    project_definition.config["name"] = project_test_data.name
    mock_create_new_proj_def = Mock(
        return_value=(project_test_data.name, project_definition)
    )
    mock_register_project = Mock()
    mock_attach_to_project_with_id = Mock()
    os_path_exists_mock = Mock(return_value=False)

    with patch.multiple(
        "anyscale.controllers.project_controller",
        create_new_proj_def=mock_create_new_proj_def,
        register_or_attach_to_project=mock_register_project,
        attach_to_project_with_id=mock_attach_to_project_with_id,
    ), patch.multiple("os.path", exists=os_path_exists_mock):
        project_controller.init(project_id=project_test_data.id, name=None)

    mock_create_new_proj_def.assert_not_called()
    mock_attach_to_project_with_id.assert_called_once_with(
        project_test_data.id, project_controller.api_client
    )
    mock_register_project.assert_not_called()


@pytest.mark.parametrize("project_id", [None, "mock-project-id"])
@pytest.mark.parametrize("project_last_cloud_id", [None, "mock-project-cloud-id"])
@pytest.mark.parametrize("org_default_cloud", [None, "mock-default-cloud-id"])
def test_write_sample_compute_config(
    mock_auth_api_client,
    project_id: Optional[str],
    project_last_cloud_id: Optional[str],
    org_default_cloud: Optional[str],
    userinfo_test_data: UserInfo,
) -> None:
    project_controller = ProjectController()

    # Set up mocks.
    project_controller.anyscale_api_client.get_project.return_value.result.last_used_cloud_id = (
        project_last_cloud_id
    )
    first_mock_cloud_record = Mock()
    first_mock_cloud_record.id = "first-mock-cloud-id"
    last_mock_cloud_record = Mock()
    last_mock_cloud_record.id = "last-mock-cloud-id"
    project_controller.api_client.list_clouds_api_v2_clouds_get.return_value.results = [
        first_mock_cloud_record,
        last_mock_cloud_record,
    ]
    mock_config = Mock()
    mock_config.to_dict = Mock(return_value={"mock-config-key": "mock-config-value"})
    project_controller.api_client.get_default_compute_config_api_v2_compute_templates_default_cloud_id_get.return_value.result = (
        mock_config
    )

    userinfo_test_data.organizations[0].default_cloud_id = org_default_cloud
    project_controller.api_client.get_user_info_api_v2_userinfo_get.return_value.result = (
        userinfo_test_data
    )

    # Run the test.
    with patch("builtins.open", mock_open_factory()) as mock_open, patch(
        "anyscale.controllers.project_controller.get_cloud_id_and_name", Mock()
    ) as mock_get_cloud_id:

        # write_sample_compute_config with default cloud.
        if org_default_cloud:
            project_controller._write_sample_compute_config(filepath="mock-fp")
            project_controller.api_client.get_default_compute_config_api_v2_compute_templates_default_cloud_id_get.assert_called_once_with(
                cloud_id=org_default_cloud
            )
            mock_get_cloud_id.assert_called_once_with(
                project_controller.api_client, cloud_id=org_default_cloud
            )
            mock_open.assert_called_once_with("mock-fp", "w")
        # write_sample_compute_config with no project ID.
        elif not project_id:
            project_controller._write_sample_compute_config(filepath="mock-fp")

            project_controller.anyscale_api_client.get_project.assert_not_called()
            project_controller.api_client.list_clouds_api_v2_clouds_get.assert_called_once_with()
            project_controller.api_client.get_default_compute_config_api_v2_compute_templates_default_cloud_id_get.assert_called_once_with(
                cloud_id=last_mock_cloud_record.id
            )
            mock_open.assert_called_once_with("mock-fp", "w")

        # write_sample_compute_config
        # with project ID, but no default cloud there.
        elif not project_last_cloud_id:
            project_controller._write_sample_compute_config(
                filepath="mock-fp", project_id=project_id
            )

            project_controller.anyscale_api_client.get_project.assert_called_once_with(
                project_id
            )
            project_controller.api_client.list_clouds_api_v2_clouds_get.assert_called_once_with()
            project_controller.api_client.get_default_compute_config_api_v2_compute_templates_default_cloud_id_get.assert_called_once_with(
                cloud_id=last_mock_cloud_record.id
            )
            mock_open.assert_called_once_with("mock-fp", "w")

        # write_sample_compute_config
        # with project ID, and with a default cloud there.
        else:
            project_controller._write_sample_compute_config(
                filepath="mock-fp", project_id=project_id
            )

            project_controller.anyscale_api_client.get_project.assert_called_once_with(
                project_id
            )
            project_controller.api_client.list_clouds_api_v2_clouds_get.assert_not_called()
            project_controller.api_client.get_default_compute_config_api_v2_compute_templates_default_cloud_id_get.assert_called_once_with(
                cloud_id=project_last_cloud_id
            )
            mock_open.assert_called_once_with("mock-fp", "w")


@pytest.mark.parametrize("name", ["", "project_a"])
@pytest.mark.parametrize("json_format", [True, False])
@pytest.mark.parametrize("any_creator", [True, False])
def test_list(mock_auth_api_client, name: str, json_format: bool, any_creator: bool):

    project_controller = ProjectController()
    mock_creator_id = "usr_abc"
    mock_userinfo_response = Mock()
    mock_userinfo_response_result = Mock()
    mock_userinfo_response.result = mock_userinfo_response_result
    mock_userinfo_response_result.id = mock_creator_id
    mock_get_user_info_api_v2_userinfo_get = Mock(return_value=mock_userinfo_response)

    project_controller.api_client.get_user_info_api_v2_userinfo_get = (
        mock_get_user_info_api_v2_userinfo_get
    )

    mock_paging_token = None
    mock_metadata = Mock()
    mock_metadata.next_paging_token = mock_paging_token

    mock_project_a = Mock()
    mock_project_a.id = "prj_a"
    mock_project_a.name = "project_a"
    mock_project_a.description = "description for project a"

    mock_project_b = Mock()
    mock_project_b.id = "prj_b"
    mock_project_b.name = "project_b"
    mock_project_b.description = "description for project b"

    mock_response = Mock()
    mock_response.metadata = mock_metadata
    mock_response.results = [mock_project_a, mock_project_b]

    # self.anyscale_api_client.search_projects(project_query)
    mock_search_projects = Mock(return_value=mock_response)
    project_controller.anyscale_api_client.search_projects = mock_search_projects

    project_controller.list(name, json_format, any_creator, max_items=20)

    if any_creator:
        project_controller.api_client.get_user_info_api_v2_userinfo_get.assert_not_called()
    else:
        project_controller.api_client.get_user_info_api_v2_userinfo_get.assert_called_once_with()

    project_controller.anyscale_api_client.search_projects.assert_called_once()
    (
        projects_query,
    ) = project_controller.anyscale_api_client.search_projects.call_args[0]
    if name:
        assert projects_query.name.equals == name
    else:
        assert projects_query.name is None
    if any_creator:
        assert projects_query.creator_id is None
    else:
        assert projects_query.creator_id.equals == mock_creator_id
    assert projects_query.paging.count == 20
    assert projects_query.paging.paging_token is None


@pytest.mark.parametrize("existing_projects", [[], [Mock()]])
def test_create_project(mock_auth_api_client, existing_projects) -> None:
    project_controller = ProjectController()
    project_controller.api_client.find_project_by_project_name_api_v2_projects_find_by_name_get = Mock(
        return_value=Mock(results=existing_projects)
    )

    project_controller.create("proj_name")

    if existing_projects:
        project_controller.api_client.create_project_api_v2_projects_post.assert_not_called()
    else:
        project_controller.api_client.create_project_api_v2_projects_post.assert_called_once_with(
            write_project={
                "name": "proj_name",
                "description": "",
                "initial_cluster_config": None,
            }
        )



import os
from typing import Optional

import click

from anyscale.controllers.cluster_controller import ClusterController
from anyscale.util import validate_non_negative_arg


HIDE_BYOD_FLAGS = os.getenv("ANYSCALE_BYOD") != "1"


@click.group("cluster", help="Interact with clusters on Anyscale.")
def cluster_cli() -> None:
    pass


@cluster_cli.command(
    name="start", help="Start or update and restart a cluster on Anyscale."
)
@click.option(
    "--name",
    "-n",
    required=False,
    default=None,
    help="Name of new or existing cluster to start or update.",
)
@click.option(
    "--env",
    required=False,
    default=None,
    help=(
        "Set the Anyscale app config to use for the cluster. This is a cluster "
        "environment name optionally followed by a colon and a build version number. "
        "Eg: my_cluster_env:1"
    ),
)
@click.option(
    "--docker",
    required=False,
    default=None,
    help=("Custom docker image name."),
    hidden=HIDE_BYOD_FLAGS,
)
@click.option(
    "--python-version",
    required=False,
    default=None,
    help=("Python version for the custom docker image."),
    hidden=HIDE_BYOD_FLAGS,
)
@click.option(
    "--ray-version",
    required=False,
    default=None,
    help=("Ray version for the custom docker image."),
    hidden=HIDE_BYOD_FLAGS,
)
@click.option(
    "--compute",
    required=False,
    default=None,
    help=("Name of cluster compute that is already registered with Anyscale."),
)
@click.option(
    "--compute-file",
    required=False,
    default=None,
    help=(
        "The YAML file of the cluster compute config to launch this cluster with. "
        "An example can be found at {website}. ".format(
            website="https://docs.anyscale.com/user-guide/configure/cluster-computes",
        )
    ),
)
@click.option(
    "--cluster-id",
    "--id",
    required=False,
    default=None,
    help=(
        "Id of existing cluster to restart. This argument can be used "
        "to interact with any cluster you have access to in any project."
    ),
)
@click.option(
    "--project-id",
    required=False,
    default=None,
    help=(
        "Override project id used for this cluster. If not provided, the Anyscale project "
        "context will be used if it exists. Otherwise a default project will be used."
    ),
)
@click.option(
    "--project",
    required=False,
    default=None,
    help=(
        "Override project name used for this cluster. If not provided, the Anyscale project "
        "context will be used if it exists. Otherwise a default project will be used."
    ),
)
@click.option(
    "--cloud-name",
    required=False,
    default=None,
    help=(
        "Name of cloud to create a default cluster compute with. If a default "
        "cloud needs to be used and this is not provided, the organization default "
        "cloud will be used."
    ),
)
@click.option(
    "--idle-timeout",
    required=False,
    help="Idle timeout (in minutes), after which the cluster is stopped. Idle "
    "time is defined as the time during which a cluster is not running a user "
    "command and does not have an attached driver. Time spent running Jupyter "
    "commands, or commands run through ssh, is still considered "
    "'idle'. -1 means no timeout. Default: 120 minutes",
    type=int,
)
@click.option(
    "--user-service-access",
    required=False,
    default=None,
    type=click.Choice(["private", "public"]),
    help=(
        "Whether user service (eg: serve deployment) can be accessed by public "
        "internet traffic. If public, a user service endpoint can be queried from "
        "the public internet with the provided authentication token. "
        "If private, the user service endpoint can only be queried from within "
        "the same Anyscale cloud and will not require an authentication token."
    ),
)
def start(
    name: Optional[str],
    env: Optional[str],
    docker: Optional[str],
    python_version: Optional[str],
    ray_version: Optional[str],
    compute: Optional[str],
    compute_file: Optional[str],
    cluster_id: Optional[str],
    project_id: Optional[str],
    project: Optional[str],
    cloud_name: Optional[str],
    idle_timeout: Optional[int],
    user_service_access: Optional[str],
) -> None:
    cluster_controller = ClusterController()
    if HIDE_BYOD_FLAGS:
        assert docker is None, "The --docker flag is not enabled for your organization"
        assert (
            python_version is None
        ), "The --python-version flag is not enabled for your organization"
        assert (
            ray_version is None
        ), "The --ray-version flag is not enabled for your organization"

    cluster_controller.start(
        cluster_name=name,
        cluster_id=cluster_id,
        cluster_env_name=env,
        docker=docker,
        python_version=python_version,
        ray_version=ray_version,
        cluster_compute_name=compute,
        cluster_compute_file=compute_file,
        cloud_name=cloud_name,
        idle_timeout=idle_timeout,
        project_id=project_id,
        project_name=project,
        user_service_access=user_service_access,
    )


@cluster_cli.command(name="terminate", help="Terminate a cluster on Anyscale.")
@click.option(
    "--name",
    "-n",
    required=False,
    default=None,
    help="Name of existing cluster to terminate.",
)
@click.option(
    "--cluster-id",
    "--id",
    required=False,
    default=None,
    help=(
        "Id of existing cluster to termiante. This argument can be used "
        "to interact with any cluster you have access to in any project."
    ),
)
@click.option(
    "--project-id",
    required=False,
    default=None,
    help=(
        "Override project id used for this cluster. If not provided, the Anyscale project "
        "context will be used if it exists. Otherwise a default project will be used."
    ),
)
@click.option(
    "--project",
    required=False,
    default=None,
    help=(
        "Override project name used for this cluster. If not provided, the Anyscale project "
        "context will be used if it exists. Otherwise a default project will be used."
    ),
)
def terminate(
    name: Optional[str],
    cluster_id: Optional[str],
    project_id: Optional[str],
    project: Optional[str],
) -> None:
    cluster_controller = ClusterController()
    cluster_controller.terminate(
        cluster_name=name,
        cluster_id=cluster_id,
        project_id=project_id,
        project_name=project,
    )


@cluster_cli.command(name="archive", help="Archive a cluster on Anyscale.")
@click.option(
    "--name",
    "-n",
    required=False,
    default=None,
    help="Name of existing cluster to archive.",
)
@click.option(
    "--cluster-id",
    "--id",
    required=False,
    default=None,
    help=(
        "Id of existing cluster to archive. This argument "
        "can be used to archive any cluster you have access to in any project."
    ),
)
@click.option(
    "--project-id",
    required=False,
    default=None,
    help=(
        "Override project id used for this cluster. If not provided, the Anyscale project "
        "context will be used if it exists. Otherwise a default project will be used."
    ),
)
@click.option(
    "--project",
    required=False,
    default=None,
    help=(
        "Override project name used for this cluster. If not provided, the Anyscale project "
        "context will be used if it exists. Otherwise a default project will be used."
    ),
)
def archive(
    name: Optional[str],
    cluster_id: Optional[str],
    project_id: Optional[str],
    project: Optional[str],
) -> None:
    cluster_controller = ClusterController()
    cluster_controller.archive(
        cluster_name=name,
        cluster_id=cluster_id,
        project_id=project_id,
        project_name=project,
    )


@cluster_cli.command(
    name="list",
    help=(
        "List information about clusters on Anyscale. By default only list "
        "active clusters in current project."
    ),
)
@click.option(
    "--name",
    "-n",
    required=False,
    default=None,
    help="Name of existing cluster to get information about.",
)
@click.option(
    "--cluster-id",
    "--id",
    required=False,
    default=None,
    help=(
        "Id of existing cluster get information about. This argument can be used "
        "to interact with any cluster you have access to in any project."
    ),
)
@click.option(
    "--project-id",
    required=False,
    default=None,
    help=(
        "Override project id used for this cluster. If not provided, the Anyscale project "
        "context will be used if it exists. Otherwise a default project will be used."
    ),
)
@click.option(
    "--project",
    required=False,
    default=None,
    help=(
        "Override project name used for this cluster. If not provided, the Anyscale project "
        "context will be used if it exists. Otherwise a default project will be used."
    ),
)
@click.option(
    "--include-all-projects",
    is_flag=True,
    default=False,
    help="List all active clusters user has access to in any project.",
)
@click.option(
    "--include-inactive",
    is_flag=True,
    default=False,
    help="List clusters of all states.",
)
@click.option(
    "--include-archived",
    is_flag=True,
    default=False,
    help=(
        "List archived clusters as well as unarchived clusters."
        "If not provided, defaults to listing only unarchived clusters."
    ),
)
@click.option(
    "--max-items",
    required=False,
    default=20,
    type=int,
    help="Max items to show in list.",
    callback=validate_non_negative_arg,
)
def list(
    name: Optional[str],
    cluster_id: Optional[str],
    project_id: Optional[str],
    project: Optional[str],
    include_all_projects: bool,
    include_inactive: bool,
    include_archived: bool,
    max_items: int,
) -> None:
    cluster_controller = ClusterController()
    cluster_controller.list(
        cluster_name=name,
        cluster_id=cluster_id,
        project_id=project_id,
        project_name=project,
        include_all_projects=include_all_projects,
        include_inactive=include_inactive,
        include_archived=include_archived,
        max_items=max_items,
    )
